#!/usr/local/bin/wish8.4

#!/bin/sh
# the next line restarts using wish \
# exec wish8.0jp "$0" "$@"

# ======================================================================
# Name: tagrin.tcl
# Version: 1.1
# Maintener: Takahashi Tetsuro / t_taka@pluto.ai.kyutech.ac.jp
# Id: $Id: tagrin.tcl,v 1.123.2.179 2006/06/30 10:36:01 tetsu Exp $
# Purpose: Tagging
# ======================================================================

# package require idrTclObject 1.0

# ToDo
# search --> data_idgĂ
# export


# ============================================================
##############################
# idrTclObjects
# copyright 2001-2002 ideaResources, LLC
# written by Salvatore Sorrentino
# this software may be used freely
# so long as the original copyright appears
# in the header and all modifications are documented.
# 
# version 1.01 fixed non-persistent objects
# --------also allows an object to destroy itself form within itself
# 
##############################

package provide idrTclObject 1.0

namespace eval idrTclObjects {

	variable objCount 0
	variable garbageList [list]
	variable parentedGarbage
	variable properties
	
	variable collection [after 5000 {idrTclObjects::collectGarbage}]
	namespace export objCount collectGarbage
	proc collectGarbage {} {
		variable garbageList
		variable collection
		after idle {}
		foreach i $garbageList {
			catch {unobject $i}
		}
		after cancel $collection
		set garbageList [list]
		set collection [after 60000 {idrTclObjects::collectGarbage}]
	}
	
	proc __setShared__ {initer obj prop master isarray com} {
		if {$com == "w"} {
			if {$isarray == {}} {
				set ${obj}::${prop} [subst \$[subst ${initer}::${prop}]]
			} else {
				set ${obj}::${prop}($isarray) [subst \$[subst ${initer}::${prop}($isarray)]]
			}
		} elseif {$com == "u"} {
			if {$isarray == {}} {
				unset ${obj}::${prop}
			} else {
				array unset ${obj}::${prop} $isarray
			}	
		}
	}
	
	proc __accessProperties__ {obj __objname__} {
		foreach prop $idrTclObjects::properties($obj) {
			namespace inscope $__objname__ variable $prop
			if {[array exists ${obj}::${prop}]} {
				foreach arr [array names ${obj}::${prop}] {
					set ${__objname__}::${prop}($arr) [subst \$[subst ${obj}::${prop}($arr)]]
				}
			} elseif {[info exists ${obj}::${prop}]} {
				set ${__objname__}::$prop [subst \$[subst ${obj}::${prop}]]
			}
			trace variable ${__objname__}::${prop} w "idrTclObjects::__setShared__ $__objname__ $obj $prop"
			trace variable ${obj}::${prop} w "idrTclObjects::__setShared__ $obj $__objname__ $prop"
			
			trace variable ${__objname__}::${prop} u "idrTclObjects::__setShared__ $__objname__ $obj $prop"
			trace variable ${obj}::${prop} u "idrTclObjects::__setShared__ $obj $__objname__ $prop"
			
		}
	}
	
	proc doUnobject {objs} {
		foreach obj $objs {
			foreach parented [array names idrTclObjects::parentedGarbage *$obj*] {
				catch {namespace delete $idrTclObjects::parentedGarbage($parented)}
				catch {array unset idrTclObjects::parentedGarbage $parented}
			}
			namespace delete $obj
		}
	}
}

######################
# start global namespace procedures


# defines an object
proc object {name body} {
	set ::idrTclObjects::procList($name) $body
}

# a utility for creating uniquely named Tk widgets
proc unique {cmd name args} {
	incr ::idrTclObjects::objCount
	eval "$cmd $name$::idrTclObjects::objCount [join $args]"
	return $name$::idrTclObjects::objCount
}

# removes a defined object from memory, including it's children
proc unobject {args} {
	set x [after idle [list idrTclObjects::doUnobject $args]]
}


		
# instantiates a defined object in the calling object, or in the global namespace (if called from the global namespce)
proc new {obj {name {}} {persist {0}}} {
	set initial $name
	incr ::idrTclObjects::objCount
	if {$name == {}} {
		set name "object$::idrTclObjects::objCount"
	}
	
	######################
	# start object procedures
	# creates new namespace and initializes its object procedures
	namespace eval [namespace current]::$name {
		variable __objname__
		# declares an exportable procedure
		set __objname__ [namespace current]
		set idrTclObjects::properties($__objname__) [list]
		set ${__objname__}::__setting__ 0
		proc property {n} {
			variable __objname__
			namespace inscope $__objname__ variable $n
			
			if {[lsearch $idrTclObjects::properties($__objname__) $n] < 0} {
				lappend idrTclObjects::properties($__objname__) $n
			}
		}
		
		
		
		
		proc public {args} {
			eval "proc $args"
			namespace export [lindex $args 0]
		}
		
		# declares an non-exportable procedure
		proc private {args} {
			eval "proc $args"
		}
		
		# imports exportable procedures from the parent object
		proc inherit {args} {
			variable __objname__
			foreach arg $args {
				namespace import [namespace parent]::$arg
				idrTclObjects::__accessProperties__ [namespace parent] $__objname__
			}
		}
		
		# imports exportable procudures from a pre-existing defined object
		proc latch {args} {
			variable __objname__
			foreach arg $args {
				namespace import ::${arg}::*
				idrTclObjects::__accessProperties__ ::${arg}
			}
		}
		
		# returns a reference to the current object or runs a procedure in the current object
		proc this {args} {
			if {$args == {}} {
				return [namespace current]
			} else {
				eval [namespace current]::$args
			}
		}
		
		# returns a reference to the parent object or runs a procedure in the parent object
		proc ancestor {args} {
			if {$args == {}} {
				return [namespace parent]
			} else {
				eval [namespace parent]::$args
			}
		}
		
		# imports exportable procudures from a non-existing defined object, and creates the defined object provides a good layer of abstraction
		proc import {args} {
			variable __objname__
			foreach arg $args {
				set x [new $arg {} 1]
				namespace import ::${x}::*
				set idrTclObjects::parentedGarbage([namespace current],$x) $x
				idrTclObjects::__accessProperties__ ::$x $__objname__
			}
		}
	}
	# end object procedures
	######################
	

	namespace eval [namespace current]::$name [list eval $::idrTclObjects::procList($obj)]
	if {$initial == {} && $persist == 0} {
		lappend idrTclObjects::garbageList $name
	}
	return $name
}

# end global namespace procedures
# ============================================================


### DefaultConf ###
proc DefaultConf {} {
    global data tagdata binding w sysconf

    # ECh̗vf̔zu (flat:1 / 2i:0)
    #   j^gĂꍇ 0悢
    set w(flat-layout) 1

    # XN[ł́Cs̍ő̕
    set data(max-char-in-line) 10000

    # XN[̃TCY
    set w(screen-height) 20
    set w(screen-width) 80

    # ~jXN[ (on:1 / off:0)
    set w(mini-screen-on) 0
    set w(mini-screen-height) 5

    # ^ȌWo͂ɏo{^gǂ(gȂ:0 / g:1)
    set w(stdout-button) 0

    # XN[ł̍s
    set w(screen-baseline) 3

    # XN[̐F
    set w(background) white
    set w(foreground) black

    # tHg
    set w(font) k14
    #set w(font) {{lr SVbN} 12 normal}

    # eLXgtHg
    set w(textfont) k14
    # set w(textfont) {{lr oSVbN} 12 normal}

    # (true/false)`̃^Oӂ鑮̃Xg
    set data(atrb-tf-list) {MȂ}

    # text`̃^Oӂ鑮̃Xg
    set data(atrb-text-list) {I}

    # g^O̍s
    set tagdata(tag-list-num) 3

    # g^ÕXg
    set tagdata(tag-list1) {t1 t2}
    set tagdata(tag-list2) {s1 s2}
    set tagdata(tag-list3) {u1 u2}

    # F鏇DXǧ̕ɂD
    set tagdata(link-order) {t1 t2 s1 s2 linked-tag neighbor-tag current-tag sel}

    # }EXɗƂɃnCCg郊Xg
    set tagdata(cand-hilight-list) {}
#     set tagdata(cand-hilight-np) {q Ɖ}
#     set tagdata(cand-hilight-@\ꑊ) {q}

    # nCCgg̕
    set tagdata(cand-hilight-width) 3

    # NƂȂ^ÕXg
    set tagdata(link-from-list) {t1 t2}

    # Ñ^ÕXg
    set tagdata(link-tag-list) {s1 s2}

    # ċAIɃnCCg
    set tagdata(hilight-current-recursively) {}

    # ċAIɃnCCg
    # ɁCʂ̃^ǑWƂȂꍇɂnCCg
    # tagdata(hilight-current-recursively)̋@\܂
    set tagdata(hilight-link-recursively) {s1}

    # LN̏ɂ݂̂ɃnCCgD
    # ftHg͖NƂănCCg邪CŎw肷邱Ƃɂ
    # LNƂĈ
    set tagdata(hilight-directed-link-list) {s1}


    # Ŏw肳ꂽNɂāCidł
    # nCCgD
    set tagdata(hilight-same-id-list) {Ɖ}

    # œ_𓖂ĂĂ^Õׂ̗^O̐F
    set tagdata(hilight-link-list-neighbor-fg) Black
    set tagdata(hilight-link-list-neighbor-bg) orangered

    # 2ȏ㗣ꂽ̃^O̐F
    set tagdata(hilight-link-list-fg) black
    set tagdata(hilight-link-list-bg) darkorange


    # NƂȂ^O̕FƔwiF
    set tagdata(link-from-color) Black
    set tagdata(link-from-background-color) red

    # gF̃Xg
    set tagdata(color-list) {Black Gray Purple Blue Green Red Pink Orange White Navy Yellow Sienna SkyBlue Magenta DarkGreen firebrick Grey}

    # e^O̐ݒ
    set tagdata(t1-status) 1
    set tagdata(t1-fcolor) LightBlue2
    set tagdata(t1-bcolor) Black
    set tagdata(t1-bind) P
    # t2
    set tagdata(t2-status) 1
    set tagdata(t2-fcolor) purple
    set tagdata(t2-bcolor) Black
    set tagdata(t2-bind) Q
    # s1
    set tagdata(s1-status) 1
    set tagdata(s1-fcolor) lightgreen
    set tagdata(s1-bcolor) Black
    set tagdata(s1-bind) p
    # s2
    set tagdata(s2-status) 1
    set tagdata(s2-fcolor) "olive drab"
    set tagdata(s2-bcolor) Black
    set tagdata(s2-bind) q
    # u1
    set tagdata(u1-status) 1
    set tagdata(u1-fcolor) yellow
    set tagdata(u1-bcolor) Black
    set tagdata(u1-bind) K
    # u2
    set tagdata(u2-status) 1
    set tagdata(u2-fcolor) orange
    set tagdata(u2-bcolor) Black
    set tagdata(u2-bind) k

#    set tagdata(u2) 1

    # t@C̃I[v
    set binding(OpenFile) Control-o

    # ㏑ۑ
    set binding(Save) Control-s

    # ㏑GNX|[g
    set binding(Export) Alt-x

    # ^Ȍ
    set binding(RemoveTag) Control-r

    # ĕ\
    set binding(Reflesh) Control-l

    # ɖ߂
    set binding(Undo) Control-z

    # tH[JXeLXgGgO
    set binding(Unfocus) Escape

    # ID
    set binding(ISearch) Control-i

    # eLXg
    set binding(TSearch) Control-f

    # I(ۑ͂Ȃ)
    set binding(exit) Control-q

    # x̕FƔwiF
    set tagdata(warnning-fg-color) Black
    set tagdata(warnning-bg-color) Pink

    # }[Nꂽ^O̔winCCgF
    set tagdata(marked-tag-color) lightgreen
    set tagdata(from-tag-color) ivory
    set tagdata(to-tag-color) pink

    # ^MO[hƕҏW[h̃gO
    set binding(ToggleEditMode) Control-t

    # eLXgւ̕ύX̕ۑ
    set binding(ResetText) Control-m

    # I[gZ[u(0:Ȃ/1:)
    set sysconf(autosave) 1

    # Xe[^X{^(0:tȂ/1:t)
    set sysconf(status_bar) 0

    # ^O{^̓(0:^Ot^/1:^O)
    set sysconf(tagbutton) 0

    # ------------------------------
    # system only

    # [h
    set w(edit-mode) 0

    # ݑΏۂƂĂ^OtH[JXO
    set w(unfocus) 0
}


# ======================================================================
# ======================================================================
#                 Don't touch under this line!!
# ======================================================================
# ======================================================================


### Main ###
proc Main {} {
    global w file tagdata screens
    wm title . "Tagrin"

    tk useinputmethods 1
    DefaultConf
    SetConf
    MakeHeader
    MakeAttribute
    MakeScreen
    SetBinding
    MakePopupMenu
    MakeStatusBar
    MakeTag
    SetTagBind
    Initialize
#    SelectImport
#     exit
}


### Object ###
# ================================================== #
#  document
# ================================================== #
object document {
    # list of record
    property records
    # integer
    property current_index
    # R[h̐
    property num
    # eLXgɃ}b`R[h index̃Xg
    property matched_record_list
    # eLXg̃CfbNX
    property matched_index

# ------------------------------
# construct --> object
# ------------------------------
    public construct {} {
	variable records
	variable current_index
	variable num
	variable matched_record_list
	variable matched_index

	set records [list]
	set current_index 0
	set num 0
	set matched_record_list [list]
	set matched_index 0
	return [this]
    }

# ------------------------------
# replace_record --> 1
# R[hXg̒̌݌Ă郌R[hCŗ^R[hƓꊷD
# ------------------------------
    public replace_record {new_record} {
	variable records
	variable current_index

	set records \
	    [lreplace $records $current_index $current_index $new_record]

	return 1
    }

# ------------------------------
# previous
# current_index̒l 1ÕR[hɃZbg
# ------------------------------
    public previous {} {
	variable current_index
	variable num
	if {$current_index == 0} {
	    set current_index [expr $num - 1]
	} else {
	    incr current_index -1
	}
    }

# ------------------------------
# next
# current_index̒l 1̃R[hɃZbg
# ------------------------------
    public next {} {
	variable current_index
	variable num
	if {$current_index == [expr $num - 1]} {
	    set current_index 0
	} else {
	    incr current_index
	}
    }

# ------------------------------
# append_record record --> list of record
# ŗ^R[h record listɒǉ
# ------------------------------
    public append_record {args} {
	variable records
	variable num
	lappend records [lindex $args 0]
	incr num
	return $records
    }


# ------------------------------
# current_id --> current id
# ݌Ă郌R[h IDԂ
# ------------------------------
    public current_id {} {
	return [[[this]::current_record]::id]
    }


# ------------------------------
# get_records --> list of record
# record listԂ
# ------------------------------
    public get_records {} {
	variable records
	return $records
    }

# ------------------------------
# current_record --> current record
# ݌Ă郌R[hԂD
# ------------------------------
    public current_record {} {
	variable current_index
	variable records
	return [lindex $records $current_index]
    }

# ------------------------------
# num --> num
# R[h̐Ԃ
# ------------------------------
    public num {} {
	variable num
	return $num
    }


# ------------------------------
# print [file-stream]
# eR[hɂāCprintĂ
# streamw肳΁C streamɏo͂
# ------------------------------
    public print {args} {
	variable records

	foreach record $records {
	    ${record}::print $args
	}
    }


# ------------------------------
# export [file-stream]
# eR[hɂāCexportĂ
# streamw肳΁C streamɏo͂
# ------------------------------
    public export {args} {
	variable records

	foreach record $records {
	    SetData
	    ${record}::export $args
	    [this]::next
	}

    }


# ------------------------------
# text_search flag -> 1(success) or 0(fail)
# textŌsCtR[h\
# ------------------------------
    public text_search {flag} {
	variable records
	variable current_index
	variable num
	variable matched_record_list
	variable matched_index
	global data
	
	set pattern $data(search-pattern)
	switch -exact $flag {
	    first {
		# ɃqbgR[h
		set matched_record_list [list]
		if { [string match $pattern "" ] } {
		    Message "Pattern is empty."
		    return 0
		}
		set start_index $current_index
		for {set sindex $start_index} {$sindex < $num} {incr sindex} {
		    if [[lindex $records $sindex]::text_match $pattern] {
			lappend matched_record_list $sindex
		    }
		}
		if {[llength $matched_record_list] == 0} {
		    Clear
		    SetStatus "Not found"
		    return 0
		} else {
		    set matched_index 0
		    set current_index [lindex $matched_record_list $matched_index]
		    SetData
		    SearchedHilight $pattern
		    SetStatus "Search..."
		    return 1
		}
	    }
	    next {
		incr matched_index
		if {$matched_index < [llength $matched_record_list)]} {
		    set current_index [lindex $matched_record_list $matched_index]
		    SetData
		    SearchedHilight $pattern
		    SetStatus "Next record"
		    return 1
		} else {
		    SetStatus "No more record"
		    incr matched_index -1
		}
	    }
	    previous {
		incr matched_index -1
		if {$matched_index >= 0} {
		    set current_index [lindex $matched_record_list $matched_index]
		    SetData
		    SearchedHilight $pattern
		    SetStatus "Next record"
		    return 1
		} else {
		    SetStatus "No more record"
		    incr matched_index
		}
	    }
	}
    }	


# ------------------------------
# idsearch -> 1(success) or 0(fail)
# ID IndexŌsCtR[h\
# ------------------------------
    public idsearch {feature val} {
	global data w
	variable current_index
	variable num

	switch -exact $feature {
	    id {
		set index 0
		foreach record [[this]::get_records] {
		    if [regexp $data(id) [${record}::id]] {
 			set current_index $index
			SetData
			return 1
		    }
		    incr index
		}
		Message "Unknown ID"
		return 0
	    }
	    index {
		SetData
		return 1
	    }

	    # ̏͂ł悤ɂȂ΂ȂȂ
	    # ʂ̎Ȃ
	    tf_attrib {
		set index 0
		foreach record [[this]::get_records] {
		    if {[${record}::get_attribute $val] == 1} {
			set current_index $index
			SetData
			return 1
		    }
		    incr index
		}
	    }

	    # ̏͂ł悤ɂȂ΂ȂȂ
	    # ʂ̎Ȃ
	    text_attrib {
		set pattern [$w(atrb).text.${val}.entry get]
		set index 0
		foreach record [[this]::get_records] {
		    if {[string match $pattern [${record}::get_attribute $val]]} {
			set current_index $index
			SetData
			return 1
		    }
		    incr index
		}
	    }
	}
    }
}



# ================================================== #
#  record
# ================================================== #
object record {
    # string( integer)
    property id
    # string
    property contents
    # array of TAG
    property tags
    # array
    property attribute
    # integer
    property max_tag_id
    # integerF݌Ă tag ID
    property current_id
    # array : attributeID ^O ID̃}bsO
    property aid2id
    # integer : V attribute  idt^Ƃ̒lD
    property newtag_id
    # list : NƂȂ^O(is_from) idXg
    property from_tag_list
    # list : NƂȂ^O(is_to) idXg
    property to_tag_list
    # list : NłNłȂ^O(is_constant) idXg
    property constant_tag_list
    # list : list of list {aid of link-from-id,  link-to-id} appendƂɁC
    # N(link-from)܂o^ĂȂƂɓD
    property link_reserved_list
    # list: temporal
    property done_list
    # list: ċAIɃnCCg^ÕXg
    property rec_hilight_tag_list
    # list: [tagname, from, to] ݑIĂ^Ȍ
    # }EXőIƂɈ߂nCCg悤ɁC
    # ܂C^uőIƂɓꏊ̃^OnCCgȂ悤
    property current_selected_tag

# ------------------------------
#     construct id --> object
# ------------------------------
    public construct {args} {
	variable id
	variable contents
	variable tags
	variable attribute
	variable max_tag_id
	variable current_id
	variable newtag_id
	variable from_tag_list
	variable to_tag_list
	variable constant_tag_list
	variable link_reserved_list
	variable done_list
	variable rec_hilight_tag_list
	variable current_selected_tag
	global data

	set id [lindex $args 0]
	set contents ""
	set max_tag_id 0
	set current_id -1
	set newtag_id 100
	set from_tag_list [list]
	set to_tag_list [list]
	set constant_tag_list [list]
	set link_reserved_list [list]
	set done_list [list]
	set rec_hilight_tag_list [list]
	set current_selected_tag [list]

	foreach atrb $data(atrb-tf-list) {
	    set attribute($atrb) 0
	}
	foreach atrb $data(atrb-text-list) {
	    set attribute($atrb) ""
	}

	return [this]
    }


# ------------------------------
# copy record --> 1
# ɗ^ꂽ record̏Rs[
# ------------------------------
    public copy {source_record} {
	variable id
	variable contents
	variable tags
	variable attribute
	variable max_tag_id
	variable current_id
	variable aid2id
	variable newtag_id
	variable from_tag_list
	variable to_tag_list
	variable constant_tag_list
	variable link_reserved_list
	variable done_list
	variable rec_hilight_tag_list
	variable current_selected_tag

	# ڃf[^ĂCobjectwɔĂ
	set id [set ${source_record}::id]
	set contents [set ${source_record}::contents]
	set max_tag_id [set ${source_record}::max_tag_id]
	set current_id [set ${source_record}::current_id]
	set newtag_id [set ${source_record}::newtag_id]
	set from_tag_list [set ${source_record}::from_tag_list]
	set to_tag_list [set ${source_record}::to_tag_list]
	set constant_tag_list [set ${source_record}::constant_tag_list]
	set link_reserved_list [set ${source_record}::link_reserved_list]
	set done_list [set ${source_record}::done_list]
	set rec_hilight_tag_list [set ${source_record}::rec_hilight_tag_list]
	set current_selected_tag [set ${source_record}::current_selected_tag]

	array set attribute [array get ${source_record}::attribute]
	array set aid2id [array get ${source_record}::aid2id]

	# ł̓^Õ|C^Rs[Ă̂ŁCVǉ
	# ^OƃN͂Ă^OɂẮĈ̂KvƂȂ
	# Ƃ΁Ctags()ŌtȂꍇ͂Ȃ̂ƎvĖȂǁD
 	array set tags [array get ${source_record}::tags]

	return 1
    }


# ------------------------------
# replace_tag tag_list
# tag_list ɓĂ^OCݎĂ^OƓꂩ
# ------------------------------
    public replace_tag {args} {
	variable tags

	set tag_list [lindex $args 0]
	foreach tag $tag_list {
	    set tags([${tag}::id]) $tag
	}
    }


# ------------------------------
# set_tag_undo
# undô߂Ƀ^ORs[
# ------------------------------
    public set_tag_undo {args} {
	global data
	variable tags
	variable aid2id
	set tag [lindex $args 0]

	if {$data(undo-flag) == 0} {
	    return 1
	}

	set index [expr [llength $data(undo-tag-list)] - 1]
	set last_list [lindex $data(undo-tag-list) $index]

	# ^ORs[ĒǉD
	set new_tag [new tag {} 1]
	${new_tag}::copy $tag
	lappend last_list $new_tag

	# NĂCÑ^ORs[D
	set from_tag_aid [${tag}::get_atrb ln]
	if ![string match $from_tag_aid ""] {

	    if [info exist aid2id($from_tag_aid)] {
		# W̃^O݂
		set from_tag_id $aid2id($from_tag_aid)
		# W֌W[vɂȂĂ^OƂɂ́C
		# WłɏĂꍇ̂ NULL check
		if {$tags($from_tag_id) != "NULL"} {
		    set new_from_tag [new tag {} 1]
		    ${new_from_tag}::copy $tags($from_tag_id)
		    lappend last_list $new_from_tag
		}
	    } else {
		# W̃^O݂Ȃ
		# 炩̗RŁCW̃^O݂Ȃꍇ
		# ̏ꍇCN̏
		${tag}::remove_atrb ln
		Message "unknown linking ID $from_tag_aid"
	    }
	}

	set data(undo-tag-list) \
	    [lreplace $data(undo-tag-list) $index $index $last_list]

    }



# ------------------------------
# reset_from_tag_list --> -1 or tagid
# SetCurrentFromɌĂ΂
# ݑIĂ from-tagpɁCfrom_tag_listȂD
# V from_tag_listlĂC̐擪ԂCĂȂ
# -1ԂD
# ------------------------------
    public reset_from_tag_list {} {
	variable current_id
	variable from_tag_list
	variable tags
	variable max_tag_id
	global tagdata

	set tmp_list [list]
	for {set i 0} {$i < $max_tag_id} {incr i} {
	    set tag $tags($i)
	    if {$tag == "NULL"} continue

	    if [${tag}::is_from] {
		if {[lsearch -exact $tagdata(current-from-list) [set ${tag}::name]] != -1} {
		    # ^OݑIĂꍇɒǉ
		    lappend tmp_list [${tag}::id]
		}
	    }
	}
	set from_tag_list $tmp_list

	if {[llength $from_tag_list] > 0} {
	    # from_tag_listlĂāC
	    if {[lsearch -exact $from_tag_list $current_id] == -1} {
		# ܂ł current_id from_tag_listɓĂȂ΁C擪̃^O idԂ
		return [lindex $from_tag_list 0]
	    } else {
		# ܂ł current_id from_tag_listɓĂ΁Ccurrent_idԂ
		return $current_id
	    }
	} else {
	    # from_tag_listlĂȂC-1ԂD
	    return -1
	}
    }

# ------------------------------
# text_match string --> 0/1
# R[h contents stringɃ}b` 1Ԃ    
# ------------------------------
    public text_match {args} {
	variable contents
	set pattern [lindex $args 0]
	return [regexp $pattern $contents]
    }


# ------------------------------
# set_tags --> 1
# SetDataɌĂ΂C^Ȍt^
# mode: String = tagging/edit
# ------------------------------
    public set_tags {mode unfocus} {
	variable current_id
	variable from_tag_list
	variable to_tag_list
	variable constant_tag_list
	variable done_list
	global tagdata

	# current-from-list
	if {${unfocus} == 1} {
	    set current_id -1
	} else {
	    set current_id [[this]::reset_from_tag_list]
	}

	# current_tag
 	set current_tag [[this]::id2tag $current_id]


	# constant^O͖Ƀ^Ot^
	foreach tag_id $constant_tag_list {
	    [[this]::id2tag $tag_id]::annotate
	}

	# from_tag_list̃^O͖Ƀ^Ot^
	foreach tag_id $from_tag_list {
	    [[this]::id2tag $tag_id]::annotate
	}

	# mode`FbN
	if {[string match $mode "edit"]} {
	    # edit modeȂCׂĂ to^Ot^
	    foreach tag_id $to_tag_list {
		[[this]::id2tag $tag_id]::annotate
	    }
	} else {
	    # tagging modeȂCt^^OI
	    if {![string match $current_tag NULL]} {

		# 郊N^ÕXg
		set current_to_tag_list [list]

		# linked tagt^
		foreach each_to_id [set ${current_tag}::linked_list] {

		    # current_tag linked_listɓo^Ăe^Oidɂ
		    set each_to_tag [[this]::id2tag $each_to_id]
		    if {[string match $each_to_tag NULL]} {
			# C̃^OȂ linked_list
			${current_tag}::delete_linked $each_to_id
		    } else {
			# ^OΕ\D
			${each_to_tag}::annotate
			lappend current_to_tag_list ${each_to_tag}
		    }
		}

		# ċAIɃ^Ot^
		set done_list [hilight_recursively $current_tag 0 [list]]
    		hilight_link_recursively $current_to_tag_list $done_list
  		set_same_id_tags ${current_tag}

		# current_tag 2ڂ̐ɃJ[\u
		ScreenCommand mark set insert [${current_tag}::from]
		ScreenCommand mark set anchor insert
		ScreenCommand mark set insert "insert + 1 chars"
		ScreenCommand mark set anchor "anchor + 1 chars"
	    }
	}

	foreach tag_name $tagdata(link-order) {
	    catch {ScreenCommand tag raise $tag_name}
	}

	return 1
    }


    # ------------------------------
    # hilight_link_recursively {$current_to_tag_list $done_list}
    # ------------------------------
    public hilight_link_recursively {current_to_tag_list $done_list} {
	variable from_tag_list
	variable done_list
	global tagdata

	# $tagdata(hilight-link-recursively)Ȃ牽Ȃ
	if { [llength $tagdata(hilight-link-recursively)] == 0 } {
	    return 1
	}

	# ̃^Õ݂R[hW߂ă`FbN
	set checking_from_tag_list [list]
	foreach tag_id $from_tag_list {
	    set a_from_tag [[this]::id2tag $tag_id]
	    if {[${a_from_tag}::is_to]} {
		if { [lsearch -exact $tagdata(hilight-link-recursively) [set ${a_from_tag}::name]] != -1 } {
		    lappend checking_from_tag_list $a_from_tag
		}
	    }
	}
	# $tagdata(hilight-link-recursively)ɓĂ
	# fromtõ^Ô݂ checking_from_tag_listɓ

	set next_to_tag_list [list]
	foreach a_fromto_tag $checking_from_tag_list {
	    if { ([lsearch -exact $done_list $a_fromto_tag]) != -1 } {
		continue
	    }
	    foreach a_to_tag $current_to_tag_list {
		if { [set ${a_fromto_tag}::id] == [set ${a_to_tag}::id] } {
		    continue
		}
		if { [set ${a_fromto_tag}::from] == [set ${a_to_tag}::from] \
			 && [set ${a_fromto_tag}::to] == [set ${a_to_tag}::to] } {
# 		    puts [${a_fromto_tag}::print]
		    # ΏۂƂȂĂ fromto_taĝC
		    # nCCgĂ a_to_tagƓ͈͂
		    # Cfromto_tagɂ
		    # hilight_recursivelyĂ
		    set done_list [hilight_recursively $a_fromto_tag 1 $done_list]
		}
	    }
	}
    }


    # ------------------------------
    # hilight_recursively($current_tag, $done_list)
    #   $current_tagċAIɃnCg
    #   $hilight_flagɏ]ănCCgF߂
    #   $done_listɓĂ^O̓nCCgȂ
    # ------------------------------
    public hilight_recursively {current_tag hilight_flag done_list} {
	variable from_tag_list
	variable to_tag_list
	variable rec_hilight_tag_list
	global tagdata

# 	puts "c: "
# 	puts ${current_tag}
# 	puts "d:"
# 	puts ${done_list}
# 	puts ""

	# ----------
	# łɃ^OnCCgĂCŏI
	if { ([lsearch -exact $done_list $current_tag]) != -1 } {
	    return $done_list
	}

	# ----------
	# nCCg
	switch $hilight_flag {
	    0 {
		# current_tagnCCg
		${current_tag}::annotate current-tag
		ScreenCommand tag configure current-tag -foreground $tagdata(link-from-color) -borderwidth 2 -underline true -background $tagdata(link-from-background-color)
		ScreenCommand tag raise current-tag
	    }
	    1 {
		# current_tagׂ̗
		set fg $tagdata(hilight-link-list-neighbor-fg)
		set bg $tagdata(hilight-link-list-neighbor-bg)
		${current_tag}::annotate neighbor-tag
		ScreenCommand tag configure neighbor-tag -foreground $tagdata(hilight-link-list-neighbor-fg) -borderwidth 2 -underline true -background $tagdata(hilight-link-list-neighbor-bg)
		ScreenCommand tag raise neighbor-tag
	    }
	    default {
		# current_tag 2ȏ㗣Ă^O
		set fg $tagdata(hilight-link-list-fg)
		set bg $tagdata(hilight-link-list-bg)
		${current_tag}::annotate linked-tag
		ScreenCommand tag configure linked-tag -foreground $tagdata(hilight-link-list-fg) -borderwidth 2 -underline true -background $tagdata(hilight-link-list-bg)
		ScreenCommand tag raise linked-tag
	    }
	}

	lappend done_list $current_tag
	set hilight_flag [expr $hilight_flag + 1]

	# ----------
	set ok 0
	# Ώۂ̃^OċAIɌJԂ`FbN
	# hilight-current-recursively ||
	# hilight-link-recursively
	# && link-from-list && link-tag-list  ͂͂2005-8-21(Sun)
	if { ([lsearch -exact $tagdata(hilight-link-recursively) [set ${current_tag}::name]] != -1) || ([lsearch -exact $tagdata(hilight-current-recursively) [set ${current_tag}::name]] != -1)} {
# 	    if { ([lsearch -exact $tagdata(link-from-list) [set ${current_tag}::name]] != -1) && ([lsearch -exact $tagdata(link-tag-list) [set ${current_tag}::name]] != -1) } {
	    set ok 1
# 	    }
	}
	if { $ok == 0 } {
	    return $done_list
	}


	# ----------
	# ċAIɌĂ
	set id [${current_tag}::get_atrb id]
	set ln [${current_tag}::get_atrb ln]

	foreach a_checking $rec_hilight_tag_list {
	    # ̂ȂΉȂ
	    if {[${a_checking}::get_atrb id] == $id} {
		continue
	    }
	    # łɃnCCgĂΉȂ
	    if {[lsearch -exact $done_list $a_checking] != -1} {
		continue
	    }
	    # W茳̃nCCg
	    if {[${a_checking}::get_atrb id] == $ln} {
		# directedłȂƂW茳ɂL΂
		# directed̂Ƃ͐L΂ȂD
		if {[lsearch -exact $tagdata(hilight-directed-link-list) [set ${current_tag}::name]] == -1} {

		    # W茳ɐL΂Ƃɂ́Chilight-link-list-neighborł
		    # ȂClink-listŃnCCg邽
		    set done_list [hilight_recursively $a_checking [expr $hilight_flag + 1] $done_list]
		}
		continue
	    }
	    # W̃nCCg
	    if {[${a_checking}::get_atrb ln] == $id} {
		set done_list [hilight_recursively $a_checking $hilight_flag $done_list]
		continue
	    }
	    # ꏊɂ^ÕnCCg
	    if { [set ${current_tag}::from] == [set ${a_checking}::from] \
		     && [set ${current_tag}::to] == [set ${a_checking}::to] } {
		set done_list [hilight_recursively $a_checking [expr $hilight_flag - 1] $done_list]

		continue
	    }
	}
#  	puts "\nr: "
#  	puts $done_list
	return $done_list

    }


# ------------------------------
# set_same_id_tags ${current_tag}
# from_tag_list̒ ŗ^ꂽ^OƓ id̃^OnCCg
# ꎞIȂ̗̂\Ȃ̂ŁC^O set_from_tags ${to_tag_list}Ɠ
# ̂gĂD
# ------------------------------
    public set_same_id_tags {current_tag} {
	variable from_tag_list
	global tagdata

	set cid [${current_tag}::get_atrb id]
	if {[string match $cid ""]} {
	    return 1
	}

	set ok 0
	foreach tag_id $from_tag_list {
	    set a_from_tag [[this]::id2tag $tag_id]
	    set aid [${a_from_tag}::get_atrb id]
	    if {![string match $cid ""] && [string match $cid $aid]} {
		${a_from_tag}::annotate refered-from-tag
		set ok 1
	    }
	}
	if { $ok == 1 } {
	    ScreenCommand tag configure refered-from-tag -foreground $tagdata(hilight-link-list-neighbor-fg) -borderwidth 2 -underline true -background $tagdata(hilight-link-list-neighbor-bg)
	    ScreenCommand tag raise refered-from-tag
	}
    }

# ------------------------------
#     set_attribute --> 1
#  SetDataɌĂ΂C̒lZbg
# ------------------------------
    public set_attribute {} {
	variable attribute
	global data

	foreach atrb $data(atrb-tf-list) {
	    set data(atrb-${atrb}) $attribute($atrb)
	}
	foreach atrb $data(atrb-text-list) {
	    set data(atrb-${atrb}) $attribute($atrb)
	}

	return 1
    }


# ------------------------------
#     display_tagcomment --> 1
#  SetDataɌĂ΂C^ORg̒l\
# ------------------------------
    public display_tagcomment {} {
	variable current_id
	global data w

 	set current_tag [[this]::id2tag $current_id]

	if {![string match $current_tag NULL]} {
	    set data(tagcomment) [${current_tag}::get_atrb comment]
	}

	$w(tagcomment).entry config -background gray85
	
	# F gray85ɂ
	# RgꂽƂɃG^[Ń^OZbgCwiɐF

	return 1
    }


# ------------------------------
#     set_tagcomment --> 1
#  ^ORg̒lZbg
# ------------------------------
    public set_tagcomment {} {
	variable current_id
	global data w

 	set current_tag [[this]::id2tag $current_id]
	regsub -all ";" $data(tagcomment) "," new_comment
	${current_tag}::set_atrb comment ${new_comment}
	$w(tagcomment).entry config -background gray95
	
	return 1
    }


# ------------------------------
# add_attribute atrb value --> 1
# record̑t^
# ------------------------------
    public add_attribute {args} {
	variable attribute

	set atrb [lindex $args 0]
	set value [lindex $args 1]

	set attribute($atrb) $value

	return 1
    }


# ------------------------------
# next_link_from_tag --> 1
# ftHgł TABɃoChĂ
# current_idɈړ
# ------------------------------
    public next_link_from_tag {} {
	variable max_tag_id
	variable current_id
	variable from_tag_list

	# idĂ^O̒Ccurrent_idI
	set current_index [lsearch -exact $from_tag_list $current_id]
	if {$current_index == [expr [llength $from_tag_list] - 1]} {
	    set current_id [lindex $from_tag_list 0]
	} else {
	    set current_id [lindex $from_tag_list [expr $current_index + 1]]
	}

# 	puts $current_id
	return 1
    }


# ------------------------------
# previous_link_from_tag --> 1
# ftHgł Shift-TABɃoChĂ
# current_idOɈړ
# ------------------------------
    public previous_link_from_tag {} {
	variable max_tag_id
	variable current_id
	variable from_tag_list

	# idĂ^O̒Ccurrent_idI
	set current_index [lsearch -exact $from_tag_list $current_id]
	if {$current_index == 0} {
	    set current_id [lindex $from_tag_list end]
	} else {
	    set current_id [lindex $from_tag_list [expr $current_index - 1]]
	}

# 	puts $current_id
	return 1
    }



# ------------------------------
# id2tag id --> tag / NULL
# idƂ $id tagԂ
# idɑΉ^OȂꍇ NULLԂ
# ------------------------------
    private id2tag {id} {
	variable tags
	variable max_tag_id

	if [info exist tags($id)] {
	    return $tags($id)
	} else {
	    # tȂCNULLԂ
	    return "NULL"
	}
    }


# ------------------------------
#     change_attribute --> 1
# ̒lύX
# ------------------------------
    public  change_attribute {atrb value} {
	variable attribute
	set attribute($atrb) $value
	return 1
    }


# ------------------------------
# get_attribute atrb --> value / 0
# ̒lԂ
# ------------------------------
    public get_attribute {atrb} {
	variable attribute

	if [info exist attribute($atrb)] {
	    return $attribute($atrb)
	} else {
	    return 0
	}
    }


# ------------------------------
#     return contents --> contents
# ------------------------------
    public contents {} {
	variable contents
	return $contents
    }

# ------------------------------
#     set_contents contents --> 1
# ------------------------------
    public set_contents {args} {
	variable contents
	set contents [lindex $args 0]
	return 1
    }

# ------------------------------
#     set_lastid lastid --> 1
# ------------------------------
    public set_lastid {args} {
	variable newtag_id
	set newtag_id [lindex $args 0]
	return 1
    }

# ------------------------------
#     append_tag tag mode --> 0:fail / 1:success
#     mode: tagging / read(from a file)
# ------------------------------
    public append_tag {args} {
	variable tags
	variable max_tag_id
	variable current_id
	variable aid2id
	variable newtag_id
	variable from_tag_list
	variable to_tag_list
	variable constant_tag_list
	variable link_reserved_list
	variable rec_hilight_tag_list
	global tagdata

	set new_tag [lindex $args 0]
	set mode [lindex $args 1]


	# idĂȂꍇCidt^D
	if [string match [${new_tag}::get_atrb id] ""] {
	    ${new_tag}::set_atrb id [format "newid%04d" $newtag_id]
	    incr newtag_id
	}

	# aid2idɓo^
	set aid [${new_tag}::get_atrb id]
	set aid2id($aid) [${new_tag}::id]


	if { ([lsearch -exact $tagdata(hilight-link-recursively) [set ${new_tag}::name]] != -1) || ([lsearch -exact $tagdata(hilight-current-recursively) [set ${new_tag}::name]] != -1)} {
	    lappend rec_hilight_tag_list $new_tag
	}
	

	# ǉ tag̎ޖ̏
	switch [set ${new_tag}::tag_type] {
	    fromto {
		# ---------- to ----------
		# to_tag_listIDǉ
		lappend to_tag_list [${new_tag}::id]

		# N̏ꍇCNɎ IDo^D
		set link_to [${new_tag}::get_atrb ln]
		# link_tag
		if [string match $link_to ""] {
		    # NĂȂꍇC
		    if [string match $mode "tagging"] {
			# ^MO
			set current_tag [[this]::id2tag $current_id]
			if {![string match $current_tag NULL]} {
			    ${new_tag}::set_atrb ln [${current_tag}::get_atrb id]
			    # ύX̂^O Undô߂ɕۑD
			    set_tag_undo ${current_tag}
			    # ݂ current_tagɁCnew_tag IDǉD
			    ${current_tag}::add_linked [set ${new_tag}::id]
			} else {
			    # Ȃ
			}
		    } else {
			# t@C̓ǂݍݎ
			# Ȃ
		    }
		} else {
		    # NĂꍇC
		    # Ñ^O new_tag IDǉ
		    if [info exist aid2id($link_to)] {
			# N悪łɓo^Ă΁C
			set link_from_tag [[this]::id2tag $aid2id($link_to)]
			if {![string match $link_from_tag NULL]} {
			    # ύX̂^O Undô߂ɕۑD
 			    set_tag_undo ${link_from_tag}
			    # ̃^O݂ linked_listɒǉ
			    ${link_from_tag}::add_linked [set ${new_tag}::id]
			} else {
			    # ̃^O݂Ȃ΁CbZ[Wo
			    Message "Tag($link_to) is not exist."
			    # todo ${new_tag}::remove_atrb ln ׂ
			}
		    } else {
			# N悪܂o^ĂȂ΁C
			lappend link_reserved_list [list $link_to [set ${new_tag}::id]]
		    }
		}



		# ---------- from ----------
		# ^OݑIĂꍇCfrom_tag_listIDǉ
		if {[lsearch -exact $tagdata(current-from-list) [set ${new_tag}::name]] != -1} {
		    lappend from_tag_list [${new_tag}::id]

		    # Vt^ from_tag current tagƂD
		    set current_id [${new_tag}::id]
		}

		# Clink-from̓o^܂ link-toɑ΂ link-fromꍇ
		# linked_listɒǉD
		foreach a_pair $link_reserved_list {
		    if [string match [lindex $a_pair 0] $aid] {
			${new_tag}::add_linked [lindex $a_pair 1]
		    }
		}



	    }

	    from {
		# ^OݑIĂꍇCfrom_tag_listIDǉ
		if {[lsearch -exact $tagdata(current-from-list) [set ${new_tag}::name]] != -1} {
		    lappend from_tag_list [${new_tag}::id]

		    # Vt^ from_tag current tagƂD
		    set current_id [${new_tag}::id]
		}

		# Clink-from̓o^܂ link-toɑ΂ link-fromꍇ
		# linked_listɒǉD
		foreach a_pair $link_reserved_list {
		    if [string match [lindex $a_pair 0] $aid] {
			${new_tag}::add_linked [lindex $a_pair 1]
		    }
		}
	    }

	    to {
		# to_tag_listIDǉ
		lappend to_tag_list [${new_tag}::id]

		# N̏ꍇCNɎ IDo^D
		set link_to [${new_tag}::get_atrb ln]
		# link_tag
		if [string match $link_to ""] {
		    # NĂȂꍇC
		    # ݂ current_tagw
		    set current_tag [[this]::id2tag $current_id]
		    if {![string match $current_tag NULL]} {
			${new_tag}::set_atrb ln [${current_tag}::get_atrb id]
			# ύX̂^O Undô߂ɕۑD
 			set_tag_undo ${current_tag}
			# ݂ current_tagɁCnew_tag IDǉD
			${current_tag}::add_linked [set ${new_tag}::id]
		    } else {
			# todo
			# Nꍇ͒ǉȂ
			# ̏ꍇ to_tag_list aid2id(),newtag_id
			# ͕ύXĂD
			# ̒l߂悢̂ȂD
			return 0
		    }
		} else {
		    # NĂꍇC
		    # Ñ^O new_tag IDǉ
		    if [info exist aid2id($link_to)] {
			# N悪łɓo^Ă΁C
			set link_from_tag [[this]::id2tag $aid2id($link_to)]
			if {![string match $link_from_tag NULL]} {
			    # ύX̂^O Undô߂ɕۑD
 			    set_tag_undo ${link_from_tag}
			    # ̃^O݂ linked_listɒǉ
			    ${link_from_tag}::add_linked [set ${new_tag}::id]
			} else {
			    # ̃^O݂Ȃ΁CbZ[Wo
			    Message "Tag($link_to) is not exist."
			    # todo ${new_tag}::remove_atrb ln ׂ
			}
		    } else {
			# N悪܂o^ĂȂ΁C
			lappend link_reserved_list [list $link_to [set ${new_tag}::id]]
		    }
		}
	    }

	    constant {
		if [${new_tag}::is_constant] {
		    # constant_tag_listIDǉ
		    lappend constant_tag_list [${new_tag}::id]
		}
	    }
	}



	# ^OǉD
	set tags([${new_tag}::id]) ${new_tag}
	incr max_tag_id

	return 1
    }


# ------------------------------
#     next_id --> max_tag_id
# max_tag_idԂD
# ------------------------------
    public next_id {} {
	variable max_tag_id
	set return_value ${max_tag_id}
	return ${return_value}
    }


# ------------------------------
#     get_a_tags --> tag
# ------------------------------
    public get_a_tag {index} {
	variable tags
	return $tags($index)
    }

# ------------------------------
#     id --> id
# ------------------------------
    public id {} {
	variable id
	return $id
    }


    # ------------------------------
    # current_selected_tagɂāCNbNƂɓ^O𕡐
    # nCCgȂ悤ɂĂ邪CNextNewData, PreviousNewDataňړ
    # ꍇ́CnCCgĂ^Oړ̂ŁCcurrent_selected_tag
    # Zbg
    # ------------------------------
    public reset_current_selected_tag {} {
	variable current_selected_tag
	set current_selected_tag [list]
    }


# ------------------------------
# set_current_id id --> 1
# current_idɒlZbg
# ------------------------------
    public set_current_id {select_id} {
	variable current_id
	variable current_selected_tag

	# todo
	# is_fromłȂC is_fromӎKv邩D
	# ̏qł͕KvȂƎvD

	set select_tag [[this]::id2tag $select_id]

	if [string match $select_tag NULL] {
	    set current_id -1
	    return 1
	}


	if {[llength $current_selected_tag] == 3} {
	    set name [lindex $current_selected_tag 0]
	    set from [lindex $current_selected_tag 1]
	    set to [lindex $current_selected_tag 2]
	    if {[${select_tag}::get_atrb name] == $name &&
		[set ${select_tag}::from] == $from &&
		[set ${select_tag}::to] == $to} {
		return 0
	    }
	}

	set current_selected_tag \
	    [list [${select_tag}::get_atrb name] \
		 [set ${select_tag}::from] \
		 [set ${select_tag}::to]]

	if {[${select_tag}::is_from]} {
	    set current_id $select_id
	}

	return 1
    }


# ------------------------------
#     remove_tag {[sel-from sel-to] already_removed_tag_list}
# J[\̃^OC邢͔͈͎w肳ꂽ^Öꕔ폜
#
# ċAIɃ^O폜Ƃɂ́Cselect_rangeɔ͈͂ł͂ȂC
# ^Ow肷D select_range̗vf̐ŔfłD
# ------------------------------
    public remove_tag {select_range already_removed_tag_list} {
	global tagdata
	variable tags
	variable current_id
	variable max_tag_id

	set removed_aid_list [list]
	set removed_tag [list]

	if {[llength $select_range] == 1} {
	    # w肳ꂽ^O̍폜sȂ
	    set tag [lindex $select_range 0]
	    lappend removed_aid_list [${tag}::get_atrb id]
	    lappend removed_tag ${tag}

	} else {
	    # Iꂽ͈͂܂̓J[\̈ʒu^O폜
	    set sel_start [lindex $select_range 0]
	    set sel_end [lindex $select_range 1]
	    set current_tag [[this]::id2tag $current_id]

	    if { [string length $sel_start] != 0 } {
		# XN[őIꂽ̈̃^O͂
		for {set i 0} {$i < $max_tag_id} {incr i} {
		    set tag $tags($i)
		    if {$tag == "NULL"} continue
		    if {[${tag}::status] && ![${tag}::read_only]} {
			# Ñ^OłȂ ܂ N悪 current_aid̎ 폜ΏۂɂȂ
			if [string match $current_tag NULL] {
			    set current_aid "NULL"
			} else {
			    set current_aid [${current_tag}::get_atrb id]
			}
			if {![${tag}::in_link_tag_list] || [${tag}::is_fromto] ||
			    [string match $current_aid [${tag}::get_atrb ln]]}  {
			    if [${tag}::remove_range $sel_start $sel_end] {
				# --- 폜[` --- #

				# 폜XgɒǉD
				lappend removed_tag ${tag}

				# ꂽ^O idĂ΃Xgɒǉ
				set aid [${tag}::get_atrb id]
				if ![string match $aid ""] {
				    lappend removed_aid_list $aid
				}
			    }
			}
		    }
		}

	    } else {
		# XN[ŃJ[\͈̂͂ɂ^O͂
		set cursor [GetScreenCursor]
		for {set i 0} {$i < $max_tag_id} {incr i} {
		    set tag $tags($i)
		    if {$tag == "NULL"} continue
		    if {[${tag}::status] && ![${tag}::read_only]} {
			# Ñ^OłȂ ܂ N悪 current_aid̎ 폜ΏۂɂȂ
			if [string match $current_tag NULL] {
			    set current_aid "NULL"
			} else {
			    set current_aid [${current_tag}::get_atrb id]
			}
			if {![${tag}::in_link_tag_list] || [${tag}::is_fromto] ||
			    [string match $current_aid [${tag}::get_atrb ln]]}  {
			    if [${tag}::on_cursor $cursor] {
				# --- 폜[` --- #

				# 폜XgɒǉD
				lappend removed_tag ${tag}

				# ꂽ^O idĂ΃Xgɒǉ
				set aid [${tag}::get_atrb id]
				if ![string match $aid ""] {
				    lappend removed_aid_list $aid
				}
			    }
			}
		    }
		}
	    }
	}


	# 폜ꂽ^O aidwĂm[hꍇCxo
	# ȂꍇCremoved_tagɓo^Ă^O

	# todo removed_tag drivenłD

	set orphan [list]
	for {set i 0} {$i < $max_tag_id} {incr i} {
	    set tag $tags($i)
	    if {$tag == "NULL"} continue
	    if { [lsearch $removed_aid_list [${tag}::get_atrb ln]] != -1} {
 		if { [lsearch $already_removed_tag_list $tag] == -1} {
		    lappend orphan ${tag}
 		}
	    }
	}

# 	puts "removed_tag"
# 	foreach i $removed_tag {
# 	    ${i}::print
# 	}
# 	puts "orphan"
# 	foreach i $orphan {
# 	    ${i}::print
# 	}


	if {[llength $orphan] > 0} {
	    foreach tag $orphan {
		${tag}::set_warnning
	    }
	    ScreenCommand tag configure warnning -foreground $tagdata(warnning-fg-color) -background $tagdata(warnning-bg-color)
	    ScreenCommand tag raise warnning
# 	    set rem [YN_Message "There are tags which depend on the removed tag."]
	    set rem [YN_Message "̃^OɃNĂ^O܂폜܂H"]
	    ScreenCommand tag delete warnning
	    if [string match $rem "yes"] {

		# ݂ȂƂȂ^O폜
		foreach rtag $removed_tag {
		    lappend already_removed_tag_list $rtag
		}
		foreach tag $orphan {
		    # $removed_tagɓĂ $removed_tag폜
		    if {[set del_index [lsearch $removed_tag $tag]] != -1} {
			set removed_tag [lreplace $removed_tag ${del_index} ${del_index}]
		    }

		    set rem_result [[this]::remove_tag [list ${tag}] $already_removed_tag_list]
		    # rŃLZꍇ́Cf 0Ԃ
		    if {$rem_result == 0} {
			return 0
		    }
		}
		# 폜ׂ^O폜
		foreach tag $removed_tag {
		    [this]::remove_tag_body ${tag}
		}
	    } else {
		return 0
	    }
	} else {
	    # 폜ׂ^O폜
	    foreach tag $removed_tag {
		[this]::remove_tag_body ${tag}
	    }
	}
	return 1
    }


# ------------------------------
# remove_tag_body
# ۂɃ^OD
# ------------------------------
    public remove_tag_body {tag} {
	variable tags
	variable aid2id
	variable from_tag_list
	variable to_tag_list
	variable constant_tag_list
	variable rec_hilight_tag_list

	# ύX̂^O Undô߂ɕۑD
	set_tag_undo ${tag}

	set id [${tag}::id]
	set tags($id) "NULL"

	# $rec_hilight_tag_listɓĂ炻
	set tmp_rec_hilight_tag_list [list]
	foreach atag $rec_hilight_tag_list {
	    if {[set ${tag}::id] != [set ${atag}::id]} {
		lappend tmp_rec_hilight_tag_list $atag
	    }
	}
	set rec_hilight_tag_list $tmp_rec_hilight_tag_list

	switch [set ${tag}::tag_type] {
	    fromto {
		# from_tag_list tag ID
		set index [lsearch $from_tag_list $id]
		set from_tag_list [lreplace $from_tag_list $index $index]
		# to_tag_list tag ID
		set index [lsearch $to_tag_list $id]
		set to_tag_list [lreplace $to_tag_list $index $index]
	    }
	    from {
		# from_tag_list tag ID
		set index [lsearch $from_tag_list $id]
		set from_tag_list [lreplace $from_tag_list $index $index]

	    }
	    to {
		# to_tag_list tag ID
		set index [lsearch $to_tag_list $id]
		set to_tag_list [lreplace $to_tag_list $index $index]
	    }
	    constant {
		# constant_tag_list tag ID
		set index [lsearch $constant_tag_list $id]
		set constant_tag_list [lreplace $constant_tag_list $index $index]
	    }
	}
    }



# ------------------------------
#     print [file-stream]
# streamw肳Ă΁C streamɏo͂
# ------------------------------
    public print {args} {
	variable id
	variable tags
	variable contents
	variable attribute
	variable newtag_id
	variable max_tag_id
	global data

	set stream [lindex $args 0]
	if [string match $stream ""] {
	    set stream stdout
	}

	puts $stream "<text id=$id>"

	puts $stream "<contents>"
	puts -nonewline $stream $contents
	puts $stream "</contents>"

	puts $stream "<attribute>"
	foreach atrb $data(atrb-tf-list) {
	    puts $stream [format "%s\t%s" ${atrb} $attribute($atrb)]
	}
	foreach atrb $data(atrb-text-list) {
	    puts $stream [format "%s\t%s" ${atrb} $attribute($atrb)]
	}
	puts $stream "</attribute>"

	puts $stream "<lastid>"
	puts $stream $newtag_id
	puts $stream "</lastid>"

	puts $stream "<tags>"
	for {set i 0} {$i < $max_tag_id} {incr i} {
	    set tag $tags($i)
	    if {$tag == "NULL"} continue
	    ${tag}::print $stream
	}
	puts $stream "</tags>"
	puts $stream "</text>"
    }


# ------------------------------
#     export [file-stream]
# streamw肳Ă΁C streamɏo͂
# ------------------------------
    public export {args} {
	variable id
	variable tags
	variable contents
	variable attribute
	variable newtag_id
	variable max_tag_id
	global data screen1

	set stream [lindex $args 0]
	if [string match $stream ""] {
	    set stream stdout
	}

	puts $stream "<text id=$id>"
	puts $stream "<contents>"
	# ---------- write tag ----------
# 	set tag_pos [list]
	for {set i 0} {$i < $max_tag_id} {incr i} {
	    set a_tag $tags($i)
	    if { $a_tag != "NULL" } {
		${a_tag}::add_mark $i
	    }
	}

	# Jn^O
	set me [$screen1 mark next 1.0]
	while {$me != ""} {
	    set new_me [$screen1 mark next $me]
	    if {[regexp {^pre::(.+)::([0-9]+)$} $me dummy tagname num]} {
		set tag [format "<%s>" $tagname]
		$screen1 insert $me $tag
		$screen1 mark unset $me
	    }
	    set me $new_me
	}

	# I^O
	set me [$screen1 mark previous end]
	while {$me != ""} {
	    set new_me [$screen1 mark previous $me]
	    if {[regexp {^post::(.+)::([0-9]+)$} $me dummy tagname num]} {
		set tag [format "</%s>" $tagname]
		$screen1 insert $me $tag
		$screen1 mark unset $me
	    }
	    set me $new_me
	}


	# -------------------------------
	puts -nonewline $stream [$screen1 get 1.0 end]
	puts $stream "</contents>"

	puts $stream "<attribute>"
	foreach atrb $data(atrb-tf-list) {
	    puts $stream [format "%s\t%s" ${atrb} $attribute($atrb)]
	}
	foreach atrb $data(atrb-text-list) {
	    puts $stream [format "%s\t%s" ${atrb} $attribute($atrb)]
	}
	puts $stream "</attribute>"
	puts $stream "</text>"
    }


# ------------------------------
# show_current_tag [file-stream]
# current_tag̏o͂
# current_tagɑ΂ăNĂ^Oo͂
# streamw肳Ă΁C streamɏo͂
# ------------------------------
    public show_current_tag {args} {
	variable tags
	variable current_id
	variable aid2id
	variable max_tag_id
	variable to_tag_list

	set stream [lindex $args 0]
	if [string match $stream ""] {
	    set stream stdout
	}

	if {$current_id == -1} {
	    puts $stream "unselected"
	    return
	}

 	set current_tag [[this]::id2tag $current_id]
	puts $stream "<current_tagset>"
	puts $stream "<from_tag>"
	${current_tag}::show
	puts $stream "</from_tag>"

	puts $stream "<to_tag>"
	set current_tagid [${current_tag}::get_atrb id]

	foreach i $to_tag_list {
	    set tag $tags($i)
	    if [string match $current_tagid [${tag}::get_atrb ln]] {
		${tag}::show
	    }
	}

	puts $stream "</to_tag>"
	puts $stream "</current_tagset>"

    }



# ------------------------------
# show_tag tag_id [file-stream]
# Ɏw肳ꂽ tag̏o͂
# tagɑ΂ăNĂ^Oo͂
# streamw肳Ă΁C streamɏo͂
# ------------------------------
    public show_tag {args} {
	variable tags
	variable aid2id
	variable max_tag_id

	set tag [lindex $args 0]
	set stream [lindex $args 1]
	if [string match $stream ""] {
	    set stream stdout
	}

	${tag}::show

	if [${tag}::is_from] {

	    # todo ̏q̓fobOpŎgĂȂC
	    # ΂ďCKv͂ȂD

	    for {set i 0} {$i < $max_tag_id} {incr i} {
		set eachtag $tags($i)
		if {$eachtag == "NULL"} continue
		if [${eachtag}::in_link_tag_list] {
		    if {[string match [${tag}::get_atrb id] [${eachtag}::get_atrb ln]]} {
			puts -nonewline $stream "  >"
			${eachtag}::show $stream
		    }
		}
	    }
	}
    }


# ------------------------------
# ҏWꂽeLXgɌ݂̃^ÖʒuXV
# ܂ contentŝ݂̂ɍXV
# ------------------------------
    public reset_text {} {
	variable tags
	variable contents
	variable from_tag_list
	variable to_tag_list
	variable constant_tag_list

	foreach tag_id [concat ${from_tag_list} ${to_tag_list} ${constant_tag_list}] {
	    $tags(${tag_id})::reset_position
	}

	set_contents [ScreenCommand get 1.0 end]
    }


}

# ================================================== #
#  tag
# ================================================== #
object tag {
    # string / ^O̖O()
    property name
    # integer / ^O ID
    property id
    # list of string / Ă鑮̃Xg
    property atrb_list
    # array / zŎ
    property atrb
    # integer.integer / ʒu
    property from
    # integer.integer / ʒu
    property to
    # string (from, to, constant) / ^O̎
    property tag_type
    # list of id / g linkĂ^O ID
    property linked_list
    # 0/1 / hilight candǂ
    property is_hilight_cand

# ------------------------------
#     construct name begin end atrb_list --> 0/object
#     success: object, fail: 0
# ------------------------------
    public construct {args} {
	global tagdata
	variable name
	variable id
	variable from
	variable to
	variable atrb_list
	variable atrb
	variable tag_type
	variable linked_list
	variable is_hilight_cand
	
	set name [lindex $args 0]
	set id [lindex $args 1]
	set from [lindex $args 2]
	set to [lindex $args 3]
	set tmp_atrb_list [lindex $args 4]
	set tag_type "constant"
	set linked_list [list]
	set is_hilight_cand 0
	if {[regexp {NULL} $from] || [regexp {NULL} $to]} {
	    return 0
	}

	# tag_typeZbg
	if {[lsearch $tagdata(link-from-list) $name] != -1} {
	    set tag_type "from"
	} 
	if {[lsearch $tagdata(link-tag-list) $name] != -1} {
	    set tag_type "to"
	}
	if {[lsearch $tagdata(link-tag-list) $name] != -1 && [lsearch $tagdata(link-from-list) $name] != -1} {
	    set tag_type "fromto"
	}

	# is_hilight_candZbg
	if {[lsearch $tagdata(cand-hilight-list) $name] != -1} {
	    set is_hilight_cand 1
	}	

	set atrb_list [list]

 	foreach i [split [string trimright $tmp_atrb_list ";"] ";"] {
	    regexp {^(.*)=(.*)$} $i dummy atribute value
	    set atrb($atribute) $value
	    lappend atrb_list $atribute
	}

	return [this]
    }



# ------------------------------
# copy tag -> 1
# ɗ^ꂽ tag̏Rs[
# ------------------------------
    public copy {source_tag} {
	variable name
	variable id
	variable from
	variable to
	variable atrb_list
	variable atrb
	variable tag_type
	variable linked_list
	variable is_hilight_cand

	set name [set ${source_tag}::name]
	set id [set ${source_tag}::id]
	set from [set ${source_tag}::from]
	set to [set ${source_tag}::to]
	set atrb_list [set ${source_tag}::atrb_list]
	set tag_type [set ${source_tag}::tag_type]
	set linked_list [set ${source_tag}::linked_list]
	set is_hilight_cand [set ${source_tag}::is_hilight_cand]
	
	array set atrb [array get ${source_tag}::atrb]

	return 1
    }


# ------------------------------
# add_linked tag_id
# ɗ^ꂽ "tag ID" linked_listɒǉD
# ------------------------------
    public add_linked {args} {
	variable linked_list

	set id [lindex $args 0]
	lappend linked_list $id
    }


# ------------------------------
# delete_linked tag_id
# ɗ^ꂽ "tag ID" linked_list폜D
# ------------------------------
    public delete_linked {args} {
	variable linked_list

	set id [lindex $args 0]
	set index [lsearch $linked_list $id]
	# linkded_list̒ idtȂꍇ index -1ƂȂ邪
	# ̏ꍇ lreplaceł͉NȂD
	set linked_list [lreplace $linked_list $index $index]
    }


# ------------------------------
#     is_same tag --> 0/1
#     ^O̖OƗ̈ lnȂ 1Ԃ
# ------------------------------
    public is_same {args} {
	variable name
	variable from
	variable to
	variable atrb

	set tag [lindex $args 0]

	if { ![string match $from [set ${tag}::from]] } {
	    return 0
	}

	if { ![string match $to [set ${tag}::to]] } {
	    return 0
	}

	if { ![string match $name [set ${tag}::name]] } {
	    return 0
	}

	if { ![string match [[this]::get_atrb ln] [${tag}::get_atrb ln]] } {
	    return 0
	}

	return 1
    }


# ------------------------------
# warnning -> 1
# xo
# ------------------------------
    public set_warnning {} {
	variable from
	variable to
	global screen tagdata

	ScreenCommand tag add warnning $from $to
    }

# ------------------------------
# on_cursor cursor/index
# 0:^OcursoȑɂC1:cursoȑɂȂ
# ------------------------------
    public on_cursor {cursor_index} {
	variable from
	variable to
	
	set cursor [Position_val $cursor_index]
	set tf [Position_val $from]
	set tt [Position_val $to]

	if {$tf <= $cursor && $cursor <= $tt} {
	    return 1
	}
	return 0
    }


# ------------------------------
# remove_range
# 1: Ă܂C0:^O͈̔͂ςȂ or ͈͂C
# ------------------------------
    public remove_range {sel_from sel_to} {
	variable from
	variable to
	
	set sf [Position_val $sel_from]
	set st [Position_val $sel_to]
	set tf [Position_val $from]
	set tt [Position_val $to]

# 	tag seldȂĂȂ
	if {$st < $tf || $tt < $sf} {
	    return 0
	}

# 	tag selĂ
	if {$sf <= $tf && $tt <= $st} {
	    return 1
	}

# 	tagꕔdȂĂƂ́Csel̏ύX
# 	  tag̍̕IĂ
	if {$sf <= $tf && $st <= $tt} {
	    set from $sel_to
	    return 0
	}

# 	  tag̉E̕IĂ
	if {$tf <= $sf && $tt <= $st} {
	    set to $sel_from
	    return 0
	}

	return 1
    }


# ------------------------------
#     add_sel --> 1
#     ^Ö̗I
# ------------------------------
    public add_sel {} {
	variable from
	variable to

 	ScreenActiveCommand tag add sel $from $to
	return 1
    }


# ------------------------------
#     annotate --> 1
#     annotate tag_name --> 1
#     ^Ot^
# ------------------------------
    public annotate {args} {
	variable name
	variable id
	variable from
	variable to
	variable atrb
	variable is_hilight_cand
	global tagdata w

	# status0̂Ƃ͕t^ȂD
	if ![[this]::status] {
	    return 0
	}

	# w肳Ă΁CŎw肳ă^Oӂ
	if {[llength $args] == 0} {
	    set tagname $name
	} else {
	    set tagname [lindex $args 0]
	}

	# ^Oo
	ScreenCommand tag add $tagname $from $to

	# "tag-${id}"ƂOł^Ot
	# ʒuꂽƂɁCʒuVɂӂ蒼
	ScreenCommand tag add tag-${id} $from $to

	# ^O{^̕t^^OnCCgD
	if [[this]::is_from] {
	    
	    # }EXCxgɃoCh
	    set fromtag [format "%s-%s" $name $id]
	    ScreenCommand tag add $fromtag $from $to
	    ScreenCommand tag bind $fromtag <Button-1> "SelectNewData $id"
	    ScreenCommand tag bind $fromtag <Control-Button-1> "SetStatus \"\""

	    $tagdata(${name}-cb) config -background $tagdata(from-tag-color)
	}

	if [[this]::is_to] {
	    $tagdata(${name}-cb) config -background $tagdata(to-tag-color)

	}
	if [[this]::is_constant] {
	    $tagdata(${name}-cb) config -background $tagdata(marked-tag-color)
	}


	# tagdata(cand-hilight-list)̃^OɂāCΏۂƂĂ^O
	# current-from-tagɓĂCnCCg
	if $is_hilight_cand {
	    set hilight_flag 0
	    foreach a_from $tagdata(cand-hilight-${name}) {
		if {[lsearch $tagdata(current-from-list) $a_from] != -1} {
		    set hilight_flag 1
		}
	    }

	    if {$hilight_flag == 1} {
		set candtag [format "%s-%s" $name $id]
		ScreenCommand tag add $candtag $from $to
		ScreenCommand tag bind $candtag <Enter> \
 		    "ScreenCommand tag configure $candtag \
                       -borderwidth $tagdata(cand-hilight-width) \
                       -relief ridge;
                     ScreenCommand tag raise $candtag"
		ScreenCommand tag bind $candtag <Leave> \
		    "ScreenCommand tag configure $candtag \
                       -relief flat"
	    }
	}


	ScreenCommand tag raise $name
	return 1
    }


# ------------------------------
# reset_position
# eLXgҏWꂽƂɁCV^Ȍꏊt^D
# ------------------------------
    public reset_position {} {
	variable id
	variable from
	variable to

	set tag_range [ScreenCommand tag ranges tag-${id}]
	if { [llength $tag_range] > 1 } {
	    set from [lindex $tag_range 0]
	    set to [lindex $tag_range 1]
	}
    }


# ------------------------------
#  id -> id
# idԂ
# ------------------------------
    public id {} {
	variable id
	return $id
    }


# ------------------------------
#    is_fromto --> 0/1
# link_from & link_to̎ 1Ԃ
# ------------------------------
    public is_fromto {} {
	variable tag_type

	if {$tag_type == "fromto"} {
	    return 1
	} else {
	    return 0
	}
    }


# ------------------------------
#    is_from --> 0/1
# link_from̎ 1Ԃ
# ------------------------------
    public is_from {} {
	variable tag_type

	if {$tag_type == "from" || $tag_type == "fromto"} {
	    return 1
	} else {
	    return 0
	}
    }


# ------------------------------
#    is_to --> 0/1
# link_tag_listɓĂ鎞 1Ԃ
# ------------------------------
    public is_to {} {
        return [[this]::in_link_tag_list]
    }


# ------------------------------
# in_link_tag_list tag --> 0/1
# tag link-tag-listɓĂƂ 1Ԃ
# ------------------------------
    public in_link_tag_list {} {
	variable tag_type

	if {$tag_type == "to" || $tag_type == "fromto"} {
	    return 1
	} else {
	    return 0
	}
    }


# ------------------------------
# is_constant --> 0/1
# tagCfrom-listɂ link_tag_listɂĂȂ 1Ԃ
# ------------------------------
    public is_constant {} {
	variable tag_type

	if {$tag_type == "constant"} {
	    return 1
	} else {
	    return 0
	}
    }


# ------------------------------
#     set_atrb atribute value --> 1
# ̒ǉsȂ
# ------------------------------
    private set_atrb {args} {
	variable atrb
	variable atrb_list

	set atribute [lindex $args 0]
	set value [lindex $args 1]
	set atrb($atribute) $value

	if {[lsearch -exact $atrb_list $atribute] == -1} {
	    lappend atrb_list $atribute
	}
	return 1
    }

# ------------------------------
#     get_atrb atribute --> value / ""
# ------------------------------
    private get_atrb {args} {
	variable atrb
	set atribute [lindex $args 0]
	if [info exist atrb($atribute)] {
	    return $atrb($atribute)
	} else {
	    return ""
	}
    }


# ------------------------------
#     remove_atrb atribute --> 0/1
# ------------------------------
    private remove_atrb {args} {
	variable atrb
	variable atrb_list

	set atribute [lindex $args 0]
	if [info exist atrb($atribute)] {
	    unset atrb($atribute)
	    set atrb_list [Lremove $atrb_list $atribute]
	    return 1
	} else {
	    return 0
	}
    }


# ------------------------------
# status --> 0/1
# return status
# O[oϐgĂ̂ŁC悭ȂD
# ------------------------------
    public status {} {
	global tagdata
	variable name

	return $tagdata(${name}-status)
    }


# ------------------------------
# set_status 0/1 --> 1
# set status
# ǃO[oϐ𒼐ڐĜŁC̃\bh͎gĂȂD
# ------------------------------
    public set_status {new_status} {
	global tagdata
	variable name

	set $tagdata(${name}-status) ${new_status}
	return 1
    }


# ------------------------------
# read_only --> 0/1
# return read_only flag. If the flag is unset, return 0
# ------------------------------
    public read_only {} {
	global tagdata
	variable name

	if ![info exist tagdata(${name}-ro)] {
	    return 0
	}
	return $tagdata(${name}-ro)
    }

# ------------------------------
#     print [file-stream]
# streamw肳Ă΁C streamɏo͂
# ------------------------------
    public print {args} {
	variable name
	variable id
	variable from
	variable to
	variable atrb_list
	variable atrb

	set stream [lindex $args 0]
	if [string match $stream ""] {
	    set stream stdout
	}

	puts -nonewline $stream [format "%s\tid:%s\t" $name $id ]
	puts -nonewline $stream "\[$from, $to\]\t"
	foreach each_atrb $atrb_list {
	    puts -nonewline $stream "$each_atrb=$atrb($each_atrb);"
	}
	puts $stream ""
    }


# ------------------------------
#     show [file-stream]
# streamw肳Ă΁C streamɏo͂
# ------------------------------
    public show {args} {
	variable name
	variable id
	variable from
	variable to
	variable atrb_list
	variable atrb
	global screen1

	set stream [lindex $args 0]
	if [string match $stream ""] {
	    set stream stdout
	}

	puts -nonewline $stream [format "%s\t%s\tid:%s\t" [$screen1 get $from $to] $name $id ]
	puts -nonewline $stream "\[$from, $to\]\t"
	foreach each_atrb $atrb_list {
	    puts -nonewline $stream "$each_atrb=$atrb($each_atrb);"
	}
	puts $stream ""
    }


# ------------------------------
# from {}
# ^O̊JnʒuԂ
# ------------------------------
    public from {} {
	variable from
	return $from
    }


    # ------------------------------
    # add_mark {num}
    # ݕ\ĂeLXgɁC}[NZbg
    # ------------------------------
    public add_mark {num} {
 	variable name
# 	variable id
 	variable from
 	variable to
 	variable atrb_list
 	variable atrb
# 	variable tag_type
# 	variable linked_list
# 	variable is_hilight_cand
	global screen1

	# status0̂Ƃ markt^ȂD
	if ![[this]::status] {
	    return 0
	}

	set atrb_pair_list [list]
	foreach each_atrb $atrb_list {
	    lappend atrb_pair_list "$each_atrb=$atrb($each_atrb)"
	}
	set atrb_string [join $atrb_pair_list ";"]

	if {$atrb_string == ""} {
	    set pre [format "pre::%s::%i" $name $num]
	} else {
	    set pre [format "pre::%s %s::%i" $name $atrb_string $num]
	}
	set post [format "post::%s::%i" $name $num]


	$screen1 mark set $pre $from
 	$screen1 mark set $post $to
    }


}

# ====================================================================== #
### Make Tag ###
proc MakeTag {} {
    global w tagdata

    for {set tagset 1} {$tagset <= $tagdata(tag-list-num)} {incr tagset} {
	if {[llength $tagdata(tag-list${tagset})] == 0} {
	    break
	}

	set each_set $w(bottom).set${tagset}
	frame ${each_set}
	pack ${each_set} -side left -padx 5 -pady 5 -anchor n

# 	set header ${each_set}.header
# 	frame $header
# 	label $header.tag -width 15 -text "Tag"
# 	label $header.color -width 9 -text "Color"
# 	label $header.bind -width 3 -text "Set"
# 	pack $header.tag -side left
# 	pack $header.color -side left -expand no -padx 10
# 	pack $header.bind -side left -expand no -padx 10
# 	pack $header -side top -expand no -fill x
	
	foreach each_tag $tagdata(tag-list${tagset}) {
	    set box_name ${each_tag}-box
	    set $box_name ${each_set}.$each_tag-box
	    MakeTagElem [set $box_name] $each_tag
	    pack [set $box_name] -side top -expand no -fill x
	}
    }
}


### Make labeled menu ###
proc MakeTagElem { frame_name name } {
    global w tagdata sysconf

    set label $name
    set key tagdata(${name}-bind)
    set val tagdata(${name}-status)

    frame $frame_name  -borderwidth 1 -relief ridge -background gray85

    frame $frame_name.name
    label $frame_name.name.label -font $w(font) -text $label
    checkbutton $frame_name.name.cb -font $w(font) -text $label -variable $val \
	-anchor w -command "ActivateTag $name"

    # annotateŃ^Ot^ƂɁC`FbN{^nCCg邽߁C
    # global variableɓĂD
    set tagdata(${name}-cb) $frame_name.name.cb

    menubutton $frame_name.menu -font $w(font) -text $tagdata(${name}-fcolor) \
	    -fg $tagdata(${name}-fcolor) -bg $tagdata(${name}-bcolor) \
	    -menu $frame_name.menu.m -relief raised -width 8
    menu $frame_name.menu.m -tearoff no
    foreach color $tagdata(color-list) {
	$frame_name.menu.m add command -label $color -command \
	  "SetColor $name ${frame_name}.menu $color"
    }	

    if {[string length [set $key]] > 0} {
 	regsub {ontrol} [set $key] "" short_name
	if {$sysconf(tagbutton) == 0} {
	    button $frame_name.entry -text ${short_name} \
 		-command "AnnTag $name button" -width 4
	} else {
	    button $frame_name.entry -text ${short_name} \
		-command "RemoveSelectedTag $name" -width 4
	}
    } else {
	button $frame_name.entry -width 10 -relief flat
    }


    if {$sysconf(status_bar) == 0} {
	pack $frame_name.name.label -side left
    } else {
	pack $frame_name.name.cb -side left
    }
    pack $frame_name.name -side left
    pack $frame_name.entry -side right -expand no
    pack $frame_name.menu -side right -expand no -padx 10
}


### SetColor ###
proc SetColor { tag_name menubutton color} {
    global tagdata

    $menubutton config -text $color
    $menubutton config -fg $color
    set tagdata(${tag_name}-fcolor) $color
    SetTagBind
    SetData
}

### ActivateTag ###
proc ActivateTag {tag_name} {
    global screen
    
    SetData    
    ScreenCommand tag raise $tag_name
}



### SetTagBind ###
# ׂẴ^OɂāCFƃL[oChsȂ
proc SetTagBind {} {
    global tagdata screens

    foreach each_tag $tagdata(tag-list) {
	ScreenCommand tag configure $each_tag -foreground $tagdata(${each_tag}-fcolor) \
	    -background $tagdata(${each_tag}-bcolor)
	if {[string length $tagdata(${each_tag}-bind)] > 0} {
	    foreach screen $screens {
		bind $screen $tagdata(${each_tag}-bind) "AnnTag $each_tag key"
	    }
	}
    }

}


### SetConf ###
proc SetConf {} {
    global tcl_platform data env tagdata

    # l̐ݒt@CJ
    switch $tcl_platform(platform) {
	unix {
	    if { [catch {source "./.tagrinrc"}] == 0 } {
		set data(config) current
	    } else {
		set tagrinrc_file [format "%s%s" $env(HOME) /.tagrinrc]
		puts $tagrinrc_file
		if { [catch {source $tagrinrc_file}] != 0 } {		
		    Message "Can't open ~/.tagrinrc \n You should save the file."
		}
		set data(config) home
	    }
	}
	windows {
	    if { [catch {source "tagrin.ini"}] != 0 } {
 		Message "Can't open tagrin.ini \n You should save the file."
	    }
	}
    }

    set tagdata(tag-list) [list]
    for {set i 1} {$i <= $tagdata(tag-list-num)} {incr i} {
	set tagdata(tag-list) [concat $tagdata(tag-list) $tagdata(tag-list${i})]
    }

}


### Initialize ###
proc Initialize {} {
    global file argc argv screen tagdata oodata

    # undolist̏
    InitialUndo

    if {$argc == 1 && [lindex $argv 0] == "-h"} {
	puts ""
	puts "tagrin.tcl \[-d input_file.xml\] \[-h\] \[input_file.tgr\]"
	puts " -h: show help and exit"
	puts " -d: output tgr file of input xml file, and exit"
	puts ""
	exit
    }

    # Ɏw肳ꂽt@CJ
    if {$argc == 1} {
	set file_name [lindex $argv 0]
	if { [catch {set file(id) [open $file_name r]}] == 0 } {
	    set file(name) $file_name
 	    ReadFile
	    SetData
	    close $file(id)
	} else {
	    Message "${file_name} couldn't opened"
	}
	return 1
    }

    if {$argc == 2 && [lindex $argv 0] == "-d"} {
	set file(import_file) [lindex $argv 1]

	if { [catch {set file(import_id) [open $file(import_file) r]}] == 0 } {
	    ImportFile
	    close $file(import_id)
	    SetData
	} else {
	    puts "ERROR: can't read input XML file."
	    exit
	}
	${oodata}::print stdout
	exit
    }


}


# -------------------- Frame --------------------

### MakePopupMenu ###
proc MakePopupMenu {} {
    global w binding
    set w(popup) .popup

    frame $w(popup) -borderwidth 2 -relief raised
#     pack $w(popup) -side top -fill x -expand no -after $w(hello)
    pack $w(popup) -side top -fill x -expand no -before $w(header)

    set file_pop $w(popup).file_pop
    set edit_pop $w(popup).edit_pop
    set view_pop $w(popup).view_pop

    menubutton $file_pop -font $w(font) -text "t@C" -menu ${file_pop}.m
    menubutton $edit_pop -font $w(font) -text "ҏW" -menu ${edit_pop}.m
    menubutton $view_pop -font $w(font) -text "\" -menu ${view_pop}.m

    pack $file_pop -side left
    pack $edit_pop -side left
    pack $view_pop -side left


    foreach i {OpenFile Save Export Reflesh Undo ISearch TSearch ResetText ToggleEditMode exit} {
	regsub {ontrol} $binding($i) "" short_${i}
    }

    menu ${file_pop}.m -font $w(font) -tearoff no
    ${file_pop}.m add command -label "J($short_OpenFile)" -command OpenFile
    ${file_pop}.m add command -label "ۑ" -command SelectSave
    ${file_pop}.m add command -label "㏑ۑ($short_Save)" -command Save
    ${file_pop}.m add command -label "C|[g" -command SelectImport
    ${file_pop}.m add command -label "GNX|[g" -command SelectExport
    ${file_pop}.m add command -label "㏑GNX|[g($short_Export)" -command Export
    ${file_pop}.m add command -label "I($short_exit)" -command exit

    menu ${edit_pop}.m -font $w(font) -tearoff no
    ${edit_pop}.m add command -label "ĕ\($short_Reflesh)" -command Reflesh
    ${edit_pop}.m add command -label "ɖ߂($short_Undo)" -command Undo
    ${edit_pop}.m add command -label "ID($short_ISearch)" -command ISearch
    ${edit_pop}.m add command -label "text($short_TSearch)" -command TSearch
    ${edit_pop}.m add command -label "[h؂ւ($short_ToggleEditMode)" -command ToggleEditMode
    ${edit_pop}.m add command -label "eLXgXV($short_ResetText)" -command ResetText

    menu ${view_pop}.m -font $w(font) -tearoff no
    ${view_pop}.m add checkbutton -label "XN[" \
	-variable w(mini-screen-on) \
	-onvalue 1 -offvalue 0 -command "SetMiniScreen"
    ${view_pop}.m add checkbutton -label "Window^O{^" \
	-variable w(tag-button-on) \
	-onvalue 1 -offvalue 0 -command "ViewTagButton"
#     ${view_pop}.m add command -label "test" -command "destroy $w(button).pre_next.pre"

    # lZbg
    set w(tag-button-on) 1
}



### ViewTagButton ###
# ^O{^ɂ
#uEBhEɂ߂ށv/uʃEBhEɕ\v̕ύX
proc ViewTagButton {} {
    global w

    if {$w(tag-button-on) == 0} {
	destroy $w(bottom)

	toplevel .tagrin_button
	wm title .tagrin_button tagrin_button

	set w(bottom) .tagrin_button.bottom
	frame $w(bottom)
	pack $w(bottom) -side top
	MakeTag
	SetData

    } else {
	destroy .tagrin_button

	set w(bottom) $w(header).bottom
	frame $w(bottom)
	pack $w(bottom) -side top
	MakeTag
	SetData
    }
}


### SetMiniScreen ###
proc SetMiniScreen {} {
    global w screen1 screen2 screens

    if {$w(mini-screen-on) == 0} {
	destroy $w(screen2)
	set screens [list $screen1]
    } else {
	destroy $w(screen1)
	destroy $w(statusbar)
	MakeScreen
	MakeStatusBar
	SetBinding
	SetTagBind
	SetData
    }
}

### ISearch ###
# IdSearch𑣂߂ɁCid index entryNA
proc ISearch {} {
    global data w

    set data(id) ""
    $w(left).index.entry delete 0 end
    Clear

    foreach i $data(atrb-tf-list) {
	$w(atrb).tf.${i} deselect
    }

    foreach i $data(atrb-text-list) {
 	$w(atrb).text.${i}.entry delete 0 end
    }
}


### TSearch ###
# screen猟sȂ
proc TSearch {} {
    global data w

    toplevel .tsearch
    wm title .tsearch Search
    label .tsearch.message -font $w(font) -text "eLXg"
    entry .tsearch.entry -font $w(font) -textvariable data(search-pattern)
    bind .tsearch.entry <Return> "TSearchBody first"

    frame .tsearch.buttons
    button .tsearch.buttons.pre -font $w(font) -text "O" -command "TSearchBody previous"
    button .tsearch.buttons.next -font $w(font) -text "" -command "TSearchBody next"
    button .tsearch.buttons.exit -text "Exit" -command "SearchExit"
    pack .tsearch.buttons.pre -side left
    pack .tsearch.buttons.next -side left
    pack .tsearch.buttons.exit -side right -padx 5
    
    pack .tsearch.message -side top -pady 2
    pack .tsearch.entry -side top -fill x -padx 10 -pady 2
    pack .tsearch.buttons -side top -pady 2
}


### SearchExit ###
# SearchI
proc SearchExit {} {
    destroy .tsearch
}

### TSearchBody ###
# TSearch̖{
proc TSearchBody {flag} {
    global oodata
    
    ${oodata}::text_search $flag
}


### SearchedHilight pattern ###
# searchf[^ɐFt
proc SearchedHilight {args} {
    global screens w
    set pattern [lindex $args 0]
    
    foreach screen $screens {
	set start_index 1.0
	while {[set searched_begin [$screen search -regexp -- $pattern $start_index end]] != ""} {
	    regexp {([0-9]+)\.([0-9]+)} $searched_begin dummy line column
	    set searched_end \
		[format "%d.%d" $line [expr $column + [string length $pattern]]]
	    $screen tag add searched  $searched_begin $searched_end
	    set start_index $searched_end
	}
	$screen tag configure searched -foreground $w(searched-fg-color) -background $w(searched-bg-color)
    }
}


### Reflesh ###
proc Reflesh {} {
    global oodata

    set record [${oodata}::current_record]
    ${record}::set_current_id -1
    SetData
    SetStatus ""
}


### InitialUndo ###
# R[hړɁCundop̃XgZbg
proc InitialUndo {} {
    global data

    set data(undo-flag) 0
    set data(undo-list) [list]
    set data(undo-tag-list) [list]
}


### SetUndoData ###
# edit operations (RemoveTag,SetTag)sтɌĂ΂C
# ݂ record data(undo-list)ɕۑ
proc SetUndoData {} {
    global data oodata

    set data(undo-flag) 1
    set current_record [${oodata}::current_record]
    set new_record [new record {} 1]
    ${new_record}::copy $current_record

    lappend data(undo-list) $new_record
    lappend data(undo-tag-list) [list]

#     puts "----saved----"
#     ${new_record}::print
}



### Undo ###
# data(undo-list) recordǂݏoClZbgD
proc Undo {} {
    global data oodata
    
    set num [llength $data(undo-list)]
    if {$num > 0} {
	set last_record [lindex $data(undo-list) [expr $num - 1]]
	set last_tag_list [lindex $data(undo-tag-list) [expr $num - 1]]

	if {$num > 1} {
	    set data(undo-list) [lrange $data(undo-list) 0 [expr $num - 2]]
	    set data(undo-tag-list) [lrange $data(undo-tag-list) 0 [expr $num - 2]]
	} else {
	    set data(undo-list) [list]
	    set data(undo-tag-list) [list]
	}

	${last_record}::replace_tag $last_tag_list
	${oodata}::replace_record $last_record

	SetData
	SetStatus "Undo!"
    } else {
	SetStatus "No more undo list."
    }
}



### MakeHeader ###
proc MakeHeader {} {
    global w data binding

    set w(header) .header
    frame $w(header) -borderwidth 2 -relief ridge
    pack $w(header) -side top -fill x -expand no -ipady 5

    set w(top) $w(header).top
    frame $w(top)
    pack $w(top) -side top -ipady 5

    set w(bottom) $w(header).bottom
    frame $w(bottom) -borderwidth 2 -relief ridge
    pack $w(bottom) -side top

    set w(left) $w(top).left
    set w(button) $w(top).right2
    set w(stdout) $w(header).out
    frame $w(left)
    frame $w(button)
    frame $w(stdout)

    if {$w(flat-layout) == 1} {
	pack $w(left) -side left -ipadx 10
	pack $w(button) -side left -fill x -expand no -padx 10
    } else {
	pack $w(left) -side top -ipadx 10
	pack $w(button) -side top -fill x -expand no -padx 10
    }

    if {$w(stdout-button) == 1} {
	pack $w(stdout) -side top -fill x -expand no -padx 10
    }
    button $w(stdout).button -text "stdout" -command OutputToSTDOUT
    pack $w(stdout).button -side top

    set filename $w(left).filename
    set id $w(left).id
    set index $w(left).index

    Makele $filename "file:" file(name) 6 15 ""
    Makele $id "ID:" data(id) 6 15 "IdSearch id dummy"

    frame $index
    label $index.label -font $w(font) -text "Index:" -width 6 -anchor w
    entry $index.entry -width 7 -font $w(font) -justify right
    label $index.slash -text "/" -font $w(font) 
    label $index.sum -font $w(font)
    pack $index.label -side left
    pack $index.entry -side left -fill x
    pack $index.slash -side left
    pack $index.sum -side left
    bind $index.entry <Return> "IdSearch index dummy"

    foreach i {filename id index} {
	pack [set $i] -side left -expand no -fill x
    }


# right2 (previous, next, remove button)
    frame $w(button).pre_next -borderwidth 2 -relief ridge
    scale $w(button).id_scale -from 0 -to 0 \
	-length 200 -orient horizontal -resolution 1
    pack $w(button).pre_next -side left -expand no
    pack $w(button).id_scale -side left -expand no -padx 10

    set pre_button $w(button).pre_next.pre
    set next_button $w(button).pre_next.next
    set remove_button $w(button).remove
    set mode_label_frame $w(button).mode
    set mode_label1 $w(button).mode.label1
    set mode_label2 $w(button).mode.label2

    regsub {ontrol} $binding(RemoveTag) "" short_Remove

    button $pre_button -text "<< Previous" -command Previous
    button $next_button -text "  Next  >> " -command Next
    button $remove_button -text "Remove ($short_Remove)" -command "RemoveTag button"
    frame $mode_label_frame

    label $mode_label1 -text "Mode:" -anchor w
    label $mode_label2 -textvariable data(mode) -anchor w
    
    pack $pre_button -side left
    pack $next_button -side left
    pack $remove_button -side left -padx 4
#     pack $mode_label_frame -side top -pady 2
    pack $mode_label1 -side left
    pack $mode_label2 -side left
}


### ResetIdScale ###
# ͂ꂽf[^ɊÂCscale𒲐D
proc ResetIdScale {} {
    global w oodata data

    set num [${oodata}::num]
    if {$num < 5} {
	set interval4 1
    } else {
	set order  [expr round(log10($num))]
	set interval1 [expr $num / 5]
	set interval2 [expr $interval1 / [Expo 10 [expr $order - 1]]]
	set interval3 [expr round($interval2)]
	set interval4 [expr $interval3 * [Expo 10 [expr $order - 1]]]
    }
    $w(button).id_scale config -to [expr $num -1] -tickinterval $interval4 \
	-command "ScaleSetData" -variable ${oodata}::current_index
#    ${oodata}::current_index̓IuWFNgwɔ

#     Index:ύX
    $w(left).index.entry config -textvariable ${oodata}::current_index
#     vύX
    $w(left).index.sum config -textvariable ${oodata}::num

#     ID:ύX
    set data(id) [${oodata}::current_id]
}

### ScaleSetData ###
# scaleĂ΂R}hDSetDataĂԁD
# scaleR}hĂԂƑɒl̂ŁC indexƂĎgD
proc ScaleSetData { index } {
    global data

    set data(id-index) $index
    SetData
}


### Expo ###
# Expo base exp
# base expԂ
proc Expo { base exp } {
    if { $exp == 0 } {
	return 1
    } else {
	return [expr $base * [Expo $base [expr $exp - 1]]]
    }
}



### MakeAttribute ###
# [Uw肵p̃`FbN{bNXƃXg{bNXD
proc MakeAttribute {} {
    global w data tagdata

    frame .under
    pack .under -side top -fill x -expand no

    set w(atrb) .under.atrb
    frame $w(atrb) -borderwidth 2 -relief ridge
    pack $w(atrb) -side top -fill x -expand no -anchor n
    set w(selfrom) .under.selfrom
    frame $w(selfrom) -borderwidth 2 -relief flat

    # Nꍇ̂݁CNĨ`FbN{^tD
    if {[llength $tagdata(link-from-list)] != 0} {
	pack $w(selfrom) -side top -padx 5 -pady 5 -fill x -expand no -anchor n
    }

    if ![info exist tagdata(current-from-list)] {
	set tagdata(current-from-list) [list [lindex $tagdata(link-from-list) 0]]
    }
    foreach from_tag $tagdata(current-from-list) {
	set tagdata([format "current-from-%s" ${from_tag}]) 1
    }

    label $w(selfrom).label -font $w(font) -text "ΏہF"
    pack $w(selfrom).label -side left -expand no -padx 2 -pady 2
    foreach tag_name $tagdata(link-from-list) {
	checkbutton $w(selfrom).cb-${tag_name} -font $w(font) -text ${tag_name} \
	    -variable tagdata(current-from-${tag_name}) \
	    -onvalue 1 -offvalue 0 -command "SetCurrentFrom"
	pack $w(selfrom).cb-${tag_name} -side left -expand no -padx 2 -pady 2
    }


    frame $w(atrb).text -borderwidth 2 -relief flat
    frame $w(atrb).tf -borderwidth 2 -relief flat

    pack $w(atrb).text -side left -fill x -expand yes -anchor n
    pack $w(atrb).tf -side left -fill x -expand yes -padx 10 -anchor n

    set data(id) "initial";

    foreach i $data(atrb-tf-list) {
 	checkbutton  $w(atrb).tf.${i} -text $i -variable data(atrb-${i}) \
 	    -font $w(font) -anchor w -command "SearchOrSetAtrb tf ${i}"
	pack $w(atrb).tf.${i} -side top -fill x
    }

    set w(tagcomment) $w(atrb).text.tagcommentxxx
    Makele $w(tagcomment) "Rg" data(tagcomment) 10 77 "SetTagComment"
    # Nꍇ̂݁CÑRgGg{bNXD
    if {[llength $tagdata(link-from-list)] != 0} {
	pack $w(tagcomment) -side top -fill x
    }

    foreach i $data(atrb-text-list) {
	Makele $w(atrb).text.${i} ${i} data(atrb-${i}) 10 77 "SearchOrSetAtrb text ${i}"
	pack $w(atrb).text.${i} -side top -fill x
    }

}



### SetCurrentFrom tag_name ###
# j[Ōݕt^^Ow肷D
proc SetCurrentFrom {} {
    global oodata
    global tagdata w
    
    set tagdata(current-from-list) [list]
    foreach tag_name $tagdata(link-from-list) {
	if {$tagdata(current-from-${tag_name}) == 1} {
	    lappend tagdata(current-from-list) $tag_name
	}
    }

    set record [${oodata}::current_record]
    set current_id [${record}::reset_from_tag_list]
    ${record}::set_current_id $current_id
    SetData
}
    

### SetTagComment ###
proc SetTagComment {} {
    global oodata

    set record [${oodata}::current_record]
    ${record}::set_tagcomment
}

### SearchOrSetAtrb ##
# type: tf / text
# atrb: selected attribute
proc SearchOrSetAtrb {type atrb} {
    global data

    # ISearch̎ $data(id)ɂ̂ŁC$data(id)Ȃ SearchƔf
    if {[string match $data(id) ""]} {
	if {[string match $type tf]} {
	    IdSearch tf_attrib $atrb
	} else {
	    IdSearch text_attrib $atrb
	}
    } else {
	SetAtrb $atrb
    }

}


### SetAtrb ###
# ͂
proc SetAtrb {atrb} {
    global data oodata
    
    set record [${oodata}::current_record]
    ${record}::change_attribute $atrb $data(atrb-${atrb})
}


### Search Data ###
# ID, Index̏ɂf[^
# SetDataĂяoCf[^\
# IdSearch feature value
proc IdSearch { feature atrb } {
    global data oodata

    AutoSave
    InitialUndo

    # id --> data(id)
    # index --> data(id-index)
    # comment --> data(atrb-Rg)

    ${oodata}::idsearch $feature $atrb
}


### Make labeled entry ###
proc Makele { name label textval lwidth ewidth command } {
    global w

    frame $name
    label $name.label -font $w(font) -text $label -width $lwidth -anchor e
#     entry $name.entry -textvariable $textval -width $ewidth -font $w(font)
    entry $name.entry -textvariable $textval -width $ewidth -font $w(font)
    pack $name.label -side left
    pack $name.entry -side left
    bind $name.entry <Return> $command
    return $name.entry
}


# ݑIĂeLXgXN[̃J[\̈ʒuԂ
proc GetScreenCursor {} {
    global w screen1 screen2

    # mini-screengĂȂ΁Cscreen1̒lԂ
    if {$w(mini-screen-on) == 0} {
	return [$screen1 index insert]
    }

#     puts "GetScreenCursor"
#     puts [$screen1 index insert]
#     puts [$screen2 index insert]
#     puts "----------"

    if [regexp screen1 [focus]] {
# 	puts "screen1!"
	return [$screen1 index insert]
    } else {
# 	puts "screen2!"
	return [$screen2 index insert]
    }
}

# ݑIĂeLXgXN[̑ÏԂ
proc GetScreenSel {} {
    global w screen1 screen2

    # mini-screengĂȂ΁Cscreen1̒lԂ
    if {$w(mini-screen-on) == 0} {
	return [$screen1 tag ranges sel]
    }

#     puts "GetScreenSel"
#     puts [$screen1 tag ranges sel]
#     puts [$screen2 tag ranges sel]
#     puts "----------"

    if [regexp screen1 [focus]] {
# 	puts "screen1!"
	return [$screen1 tag ranges sel]
    } else {
# 	puts "screen2!"
	return [$screen2 tag ranges sel]
    }
}

# ANeBuȃXN[ɑ΂ďsȂ
proc ScreenActiveCommand { args } {
    global screen1 screen2 w

    # mini-screengĂȂ΁Cscreen1̒lԂ
    if {$w(mini-screen-on) == 0} {
	eval "$screen1 $args"
    }

    if [regexp screen1 [focus]] {
	eval "$screen1 $args"
    } else {
	eval "$screen2 $args"
    }
}


# ̃XN[ɑ΂ďsȂ
proc ScreenCommand { args } {
    global screen1 screen2 w

    set retval [eval "$screen1 $args"]
    if {$w(mini-screen-on) == 1} {
	eval "$screen2 $args"
    }
    return $retval
}

### Make screen ###
proc MakeScreen {} {
    global screens w

    set screens [list]
    if {$w(mini-screen-on) == 1} {
	MakeScreen2
    }
    MakeScreen1
}


### Make screen1 ###
proc MakeScreen1 {} {
    global w screen1 screens
    set w(screen1) .screen1

    frame $w(screen1) -borderwidth 2
    pack $w(screen1) -side top -fill both -expand yes

    # List Box,Scroll Bar
    set mtext $w(screen1).mtext
    set myscroll $w(screen1).myscroll
    set msbottom $w(screen1).bottom
    set mspad $w(screen1).pad

    set screen1 [text $mtext \
		    -spacing2 $w(screen-baseline) \
		    -font $w(textfont) \
		    -width $w(screen-width) \
		    -height $w(screen-height) \
		    -background $w(background) \
		    -foreground $w(foreground) \
		    -insertbackground $w(foreground) \
		    -borderwidth 2 \
		    -relief sunken \
		    -setgrid true \
		    -yscrollcommand "$myscroll set"]
    scrollbar $myscroll -orient vertical -command {$screen1 yview}
    frame $msbottom
    set pad [expr 2 * ([$myscroll cget -bd] + \
	    [$myscroll cget -highlightthickness])]
    frame $mspad -width $pad -height $pad
    pack $mspad -in $msbottom -side right
    pack $myscroll -side right -fill y -expand no
    pack $screen1 -side left -fill both -expand yes
    lappend screens $screen1
}

### Make screen2 ###
proc MakeScreen2 {} {
    global w screen2 screens
    set w(screen2) .screen2

    frame $w(screen2) -borderwidth 2
    pack $w(screen2) -side top -fill both -expand no

    # List Box,Scroll Bar
    set mtext $w(screen2).mtext
    set myscroll $w(screen2).myscroll
    set msbottom $w(screen2).bottom
    set mspad $w(screen2).pad

    set screen2 [text $mtext \
		    -spacing2 $w(screen-baseline) \
		    -font $w(textfont) \
		    -width $w(screen-width) \
		    -height $w(mini-screen-height) \
		    -background $w(background) \
		    -foreground $w(foreground) \
		    -insertbackground $w(foreground) \
		    -borderwidth 2 \
		    -relief sunken \
		    -setgrid true \
		    -yscrollcommand "$myscroll set"]
    scrollbar $myscroll -orient vertical -command {$screen2 yview}
    frame $msbottom
    set pad [expr 2 * ([$myscroll cget -bd] + \
	    [$myscroll cget -highlightthickness])]
    frame $mspad -width $pad -height $pad
    pack $mspad -in $msbottom -side right
    pack $myscroll -side right -fill y -expand no
    pack $screen2 -side left -fill both -expand yes
    lappend screens $screen2
}

# MakeStatusBar
# make status bar
proc MakeStatusBar {} {
    global w
    set w(statusbar) .statusbar

    frame $w(statusbar) -borderwidth 2 -relief sunken 
    pack $w(statusbar) -side top -fill x

    set status_bar_head $w(statusbar).status_bar_head
    set status_bar_val $w(statusbar).status_bar_val

    label $status_bar_head -text "Status: " -fg darkred -anchor w
    label $status_bar_val -textvariable w(status) -anchor w
    pack $status_bar_head -side left
    pack $status_bar_val -side left

    SetStatus ""
}



# -------------------- command -------------------- #

### SetStatus ###
# set status information into status bar
proc SetStatus {input} {
    global w

    set w(status) $input
}


### open file ###
proc OpenFile {} {
    global file

    set ftypes {
	{"tgr" {.tgr}}
	{"Other" *}
    }
    set file_name [tk_getOpenFile -filetypes $ftypes]

    if { [catch {set file(id) [open $file_name r]}] == 0 } {
	set file(name) $file_name
	ReadFile
	SetData
	close $file(id)
    }
}


### read from file ###
proc ReadFile {} {
    global file oodata w

    SetStatus "Reading..."
    set oodata [[new document {} 1]::construct]

    while { [gets $file(id) line] >= 0 } {
	if { [regexp {^<text id=(.+)>$} $line dummy id ] } {
 	    set rec [[new record {} 1]::construct $id]

	    set buffer ""
	    
	    gets $file(id) in_line
	    while { ![regexp {^</text>} $in_line] } {
		switch -regexp -- $in_line {
		    {^#} {
		    }
		    {^<tags>} {
			set in_line [gets $file(id)]
			while { ![regexp {^</tags>} $in_line] } {
			    set tag_info [split $in_line "\t"]
			    set tag_name [lindex $tag_info 0]
			    set tag_range [lindex $tag_info 2]
			    regexp {\[([0-9.]+), ([0-9.]+)\]} $in_line dummy from_value to_value
			    set atrb_list [lindex $tag_info 3]

# 			    puts $tag_info
# 			    puts [lindex $tag_info 0]
# 			    puts [lindex $tag_info 1]
# 			    puts "$from_value, $to_value"
#  			    puts $atrb_list

			    set atag [[new tag {} 1]::construct \
					  $tag_name \
					  [${rec}::next_id] \
					  $from_value \
					  $to_value \
					  $atrb_list]
			    if {$atag != 0} {
				# ^Oǉ
				${rec}::append_tag ${atag} read
			    }
			    set in_line [gets $file(id)]
			}
		    }
		    {^<contents>} {
			set in_line [gets $file(id)]
			while { ![regexp {^</contents>} $in_line] } {
			    set buffer [format "%s%s\n" $buffer $in_line]
			    set in_line [gets $file(id)]
			}
			${rec}::set_contents $buffer
		    }
		    {^<attribute>} {
			set in_line [gets $file(id)]
			while { ![regexp {^</attribute>} $in_line] } {
			    set atrb_set [split [string trim $in_line]]
			    set atribute [lindex $atrb_set 0]
			    set value [join [lrange $atrb_set 1 end]]
			    set data(${id}-$atribute) $value
			    ${rec}::add_attribute $atribute $value
			    set in_line [gets $file(id)]
			}
		    }
		    {^<lastid>} {
			set in_line [gets $file(id)]
			while { ![regexp {^</lastid>} $in_line] } {
			    ${rec}::set_lastid $in_line
			    set in_line [gets $file(id)]
			}
			${rec}::set_contents $buffer
		    }

		}
		gets $file(id) in_line
	    }
	    ${oodata}::append_record ${rec} read
	}
    }

#    ${oodata}::print
    ResetIdScale
    SetStatus "Reading... done."
}


### SelectImport ###
proc SelectImport {} {
    global file

    if [info exist file(import_file)] {
	set file_import_file $file(import_file)
    }
    set ftypes {
	{"xml" {.xml}}
	{"Other" *}
    }
    set file(import_file) [tk_getOpenFile -filetypes $ftypes]
#    set file(import_file) "sample.xml"
    
    if { [catch {set file(import_id) [open $file(import_file) r]}] == 0 } {
	ImportFile
	close $file(import_id)
	SetData
    }
}


### ImportFile ###
proc ImportFile {} {
    global file tagdata stack oodata

    SetStatus "Import..."
    set oodata [[new document {} 1]::construct]

    while { [gets $file(import_id) line] >= 0 } {
	if { [regexp {^<text (.+)>$} $line dummy attribs ] } {
	    set attrib_list [list]
	    foreach attrib_set [split $attribs ";"] {
		regexp {(.+)=(.+)$} $attrib_set dummy attrib val
		if [string match $attrib "id"] {
		    set rec [[new record {} 1]::construct $val]
		}
		lappend attrib_list [list $attrib $val]
	    }

	    foreach attrib_set ${attrib_list} {
		if [string match [lindex $attrib_set 0] "lastid"] {
		    ${rec}::set_lastid [lindex $attrib_set 1]
		} else {
		    ${rec}::add_attribute [lindex $attrib_set 0] [lindex $attrib_set 1]
		}
	    }

	    set buffer ""
	    set line_num 1

	    gets $file(import_id) in_line
 	    # -------------------- sǂݍ -------------------- #
	    while { ![regexp {^</text>} $in_line] } {
		set buffer [format "%s%s\n" $buffer $in_line]
		set offset 0
		set line_buffer $in_line

		# 
		foreach each_tag $tagdata(tag-list) {
		    NewStack $each_tag
		}
		
		while { [regexp {<[^>]+>} $line_buffer tag_string] } {
#  		    puts "in: $tag_string"
		    set tag_begin [string first $tag_string $line_buffer]
		    regsub {<[^>]+>} $line_buffer "" line_buffer
		    set tag_name [lindex [split [string trim $tag_string {< > /}] " "] 0]
		    set tag_atrb [string trim $tag_string {< > /}]

		    if {[regexp {</} $tag_string ]} {
			# I^Ȍꍇ
			set pop_value [Pop $tag_name]
			set from_value [lindex [split $pop_value ":"] 0]
			set to_value [format "%d.%d" $line_num [expr $tag_begin]]
			set tag_value [lindex [split $pop_value ":"] 1]
			if { [string first " " $tag_value] != -1 } {
			    set atrb_list [string range $tag_value [expr [string first " " $tag_value] + 1] end]
			} else {
			    set atrb_list ""
			}

			set atag [[new tag {} 1]::construct \
				      $tag_name \
				      [${rec}::next_id] \
				      $from_value \
				      $to_value \
				      $atrb_list]
			if {$atag != 0} { ${rec}::append_tag ${atag} read}

		    } else {
			# Jn^ȌꍇCX^bNɓD
			Push $tag_name [format "%d.%d:%s" $line_num [expr $tag_begin] $tag_atrb]
		    }
		}
		# -------------------- sǂݍޏI -------------------- #
		set line_num [expr $line_num + 1]
		gets $file(import_id) in_line
	    }

	    # --- 1R[h̓ǂݍݏI
 	    regsub -all {<[^>]+>} $buffer "" new_buffer
	    ${rec}::set_contents $new_buffer
	    ${oodata}::append_record ${rec}

	}
    }

#     ${oodata}::print
    ResetIdScale
    SetStatus "Import... done."
}


### Lremove ###
# Lremove $list $val
# $list $valXgԂ
proc Lremove {list val} {
    set ret_val [list]

    foreach i $list {
	if { ![string match $i $val] } {
	    lappend ret_val $i
	}
    }
    return $ret_val
}


### Position_val ###
# int.int`̃CfbNXƂĎ󂯎CɓWJĕԂ
proc Position_val {position} {
    global data
    regexp {([0-9]+)\.([0-9]+)} $position dummy line column
    return [expr $line * $data(max-char-in-line) + $column]
}


# X^bN̎g
# z stack global
# NewStack stack_name
### NewStack ###
proc NewStack {stack_name} {
    global stack

    set stack($stack_name) [list]
}

### Push ###
proc Push {stack_name val} {
    global stack tagdata

    if {[lsearch $tagdata(tag-list) $stack_name] == -1} {
	return 0;
    }
    return [set stack($stack_name) [linsert $stack($stack_name) end $val]]
}

### Pop ###
proc Pop {stack_name} {
    global stack

    set length [llength $stack($stack_name)]

    switch $length {
	0 {
	    set ret_val NULL
	}
	1 {
	    set ret_val [lindex $stack($stack_name) end]
	    set stack($stack_name) [list]
	}
	default {
	    set ret_val [lindex $stack($stack_name) end]
	    set stack($stack_name) [lrange $stack($stack_name) 0 [expr $length - 2]]
	}
    }
    return $ret_val
}


### set data ###
# \
proc SetData {} {
    global screen tagdata w oodata data screen1 screen2

    if ![info exist oodata] { return 0 }

#     set time_start [clock clicks -milliseconds]

#     set w(position) [ScreenCommand index insert]
    set w(scroll1) [$screen1 yview]
    if {$w(mini-screen-on) == 1} {
	set w(scroll2) [$screen2 yview]
    }

    Clear
    # searched hilightɖ߂
    ScreenCommand tag delete searched
    SetStatus ""
    set record [${oodata}::current_record]
    set data(id) [${oodata}::current_id]

    # ^O{^̔wĩnCCg
    foreach tagname $tagdata(tag-list) {
	$tagdata(${tagname}-cb) config -background gray85
    }

    #--------------------------------------------------
    # ܂ŏ

    # eLXg̕\
    Screen [${record}::contents]

    # ^O̕t^
    if {$w(edit-mode) == 1} {
	set edit_mode edit
    } else {
	set edit_mode tagging
    }
    if {$w(unfocus) == 1} {
	set unfocus 1
	set w(unfocus) 0
    } else {
	set unfocus 0
    }
    ${record}::set_tags ${edit_mode} ${unfocus}



    # ̕t^
    ${record}::set_attribute
    # ^ORg̕t^
    ${record}::display_tagcomment

#     ${record}::print

#     tk::TextSetCursor ScreenCommand ${w(position)}
    $screen1 yview moveto [lindex $w(scroll1) 0]
    if {$w(mini-screen-on) == 1} {
	$screen2 yview moveto [lindex $w(scroll2) 0]
    }


#     set time_end [clock clicks -milliseconds]
#     puts [format "- time SetData: %d ms" [expr $time_end - $time_start]]
}


### $tagdata(link-from)^OIDƂāC̃^OnCCg
proc SelectNewData {newid} {
    global oodata

    set record [${oodata}::current_record]
    set ret_val [${record}::set_current_id $newid]

    if {$ret_val == 1} {
	SetData
    }
}


### NextNewData ###
# Ñ^OɈړ
proc NextNewData {} {
    global oodata

    if ![info exist oodata] { return 0 }

    set record [${oodata}::current_record]
    ${record}::next_link_from_tag
    ${record}::reset_current_selected_tag
    SetData
    return -code break 0
}


### PreviousNewData ###
# Ñ^OOɈړ
proc PreviousNewData {} {
    global oodata

    if ![info exist oodata] { return 0 }

    set record [${oodata}::current_record]
    ${record}::previous_link_from_tag
    ${record}::reset_current_selected_tag
    SetData
    return -code break 0
}


### Usort ###
# sortāCuniqƂ
proc Usort {input_list} {
    set new_list [lsort $input_list]
    set ret_list [list]
    set before "NULL"
    foreach i $new_list {
	if {$i != $before} {
	    lappend ret_list $i
	    set before $i
	}
    }
    return $ret_list
}


### next data ###
proc Next {} {
    global data oodata

    if ![info exist oodata] {
	return 0
    }

    # NextOɁC݂ attribute̒lLD
    foreach i $data(atrb-text-list) {
	SetAtrb ${i}
    }

    AutoSave
    InitialUndo

    ${oodata}::next
    SetData
}

### previous data ###
proc Previous {} {
    global data oodata

    if ![info exist oodata] {
	return 0
    }

    # PreviousOɁC݂ attribute̒lLD
    foreach i $data(atrb-text-list) {
	SetAtrb ${i}
    }

    AutoSave
    InitialUndo

    ${oodata}::previous
    SetData
}



### Clear ###
proc Clear {} {
    global screen

    ScreenCommand delete 1.0 end
}



### RemoveTag ###
# J[\̉܂͔͈͎w肳Ă^Ö̗폜
proc RemoveTag {key_or_button} {
    global tagdata oodata

    SetUndoData

    set record [${oodata}::current_record]
    ${record}::remove_tag [GetScreenSel] [list]

    selection clear
    SetData

    if {[string compare $key_or_button key] == 0} {
	return -code break 0
    }
}


### RemoveSelectedTag ###
# I^O폜
proc RemoveSelectedTag {name} {
    global tagdata oodata

    SetUndoData

    foreach tag $tagdata(tag-list) {
	set tmp($tag) $tagdata(${tag}-status)
	set tagdata(${tag}-status) 0
    }
    set tagdata(${name}-status) 1

    set record [${oodata}::current_record]
    ${record}::remove_tag [GetScreenSel] [list]

    foreach tag $tagdata(tag-list) {
	set tagdata(${tag}-status) $tmp($tag) 
    }

    selection clear
    SetData
}



### Print to screen ###
proc Screen { monstr } {
    ScreenCommand insert end $monstr
#     ScreenCommand insert end "\n"
}


### AnnTag ###
# I͈͂Ƀ^Oӂ
# eLXgɑIꂽ̈悪 SetTagĂсCłȂ
# CopyTagĂ
proc AnnTag { tag_name key_or_button } {

    SetUndoData

#     set time_start [clock clicks -milliseconds]

    if {[string length [GetScreenSel]] > 0} {
	SetTag $tag_name
    } else {
	CopyTag $tag_name
    }

#     set time_end [clock clicks -milliseconds]
#     puts [format "- time AnnTag: %d ms" [expr $time_end - $time_start]]

    if {[string compare $key_or_button key] == 0} {
	return -code break 0
    }
}
 

### Set Tag ###
# I͈͂Ƀ^Oӂ
proc SetTag { tag_name } {
    global tagdata oodata w

    if {$w(edit-mode) == 1} {
	Message "Please change the mode to \"Tagging mode\"."
	return 1
    }

    set record [${oodata}::current_record]
    if {$tagdata(${tag_name}-status) == 1} {
	catch { ScreenCommand tag add $tag_name sel.first sel.last }
	ScreenCommand tag raise $tag_name

	set select_range [GetScreenSel]
	set sel_from [lindex $select_range 0]
	set sel_to [lindex $select_range 1]

	set atag [[new tag {} 1]::construct $tag_name [${record}::next_id] $sel_from $sel_to]
	${record}::append_tag ${atag} tagging

 	SetData
    }

    selection clear
}


### CopyTag ###
# |C^ꏊɂ^Ö̗Ɏw肵^O͂
proc CopyTag { tag_name } {
    global tagdata oodata

    if {$tagdata(${tag_name}-status) == 1} {

	# todo
	# oncursorɂȂ^OꍇC UNION
	# ̗̈Ƀ^OӂׂD

	set record [${oodata}::current_record]
	
	for {set i 0} {$i < [${record}::next_id]} {incr i} {
	    set tag [${record}::get_a_tag $i]

	    if ![string match $tag NULL] {
		set cursor [GetScreenCursor]
		if [${tag}::on_cursor $cursor] {
		    ${tag}::add_sel
		    SetTag $tag_name
		    break
		}
	    }
	}
    }
}


### SelectSave ###
proc SelectSave {} {
    global file

    if [info exist file(name)] {
	set file_name $file(name)
    }
    set ftypes {
	{"sample" {.tgr}}
	{"Other" *}
    }
    set file(name) [tk_getSaveFile -defaultextension ".tgr" -filetypes $ftypes]
    if { [string length $file(name)] == 0 } {
	if [info exist file(name)] {
	    set file(name) $file_name
	}
    } else {
	Save
    }
}


### AutoSave ###
# Previous, Next, IdSearchsƁCȌԂۑD
proc AutoSave {} {
    global file data sysconf

    # todo
    # SaveƂɁCbackt@C͏

    if {$sysconf(autosave) == 1} {
	# undo_stackāClĂ Save
	set num [llength $data(undo-list)]
	if {$num > 0} {
	    set file(autosave) "$file(name)_back"
	    Save
	    set file(autosave) ""
	    # 	puts "auto saved"
	} else {
	    # 	puts "empty"
	}
    }
}

### Save ###
proc Save {} {
    global data file oodata

    # SaveOɁC݂ attribute̒lLD
    foreach i $data(atrb-text-list) {
	SetAtrb ${i}
    }

    if {[info exist file(autosave)] && [string length $file(autosave)] > 0} {
	set filename $file(autosave)
    } else {
	set filename $file(name)
    }

    if { [catch {set file(id) [open $filename w]}] == 0 } {
	SetStatus "Saving..."
	${oodata}::print $file(id)
	close $file(id)
	SetStatus "Saving... done."
    } else {
	Message "Can't Save"
    }
}


### SelectExport ###
proc SelectExport {} {
    global file
    
    if [info exist file(export_file)] {
	set file_export_file $file(export_file)
    }
    set ftypes {
	{"XML" {.xml}}
	{"Other" *}
    }
    set file(export_file) [tk_getSaveFile -defaultextension ".xml" -filetypes $ftypes]
    if { [string length $file(export_file)] == 0 } {
	if [info exist file_export_file] {
	    set file(export_file) $file_export_file
	}
    } else {
	Export
    }
}


### Export ###
proc Export {} {
    global data screen1 tagdata file oodata

    # ExportOɁC݂ attribute̒lLD
    foreach i $data(atrb-text-list) {
	SetAtrb ${i}
    }

#     set file(export_file) "test.xml"
    if { [catch {set file(export_id) [open $file(export_file) w]}] == 0 } {
	SetStatus "Exporting..."
	${oodata}::export $file(export_id)
	close $file(export_id)
	SetStatus "Exporting... done."
    } else {
	Message "Can't Export"
    }

    set record [${oodata}::current_record]
    ${record}::set_current_id -1
    SetData
}


### ToggleEditMode ###
# EditModeɓ
proc ToggleEditMode {} {
    global w

    if {$w(edit-mode) == 1} {
	ResetText
	set w(edit-mode) 0
	SetData
	SetStatus "Tagging mode"
    } else {
	set w(edit-mode) 1
	SetData
	SetStatus "Edit mode"
    }
}



### ExistMatch ###
# ExistMatch str1 str2
# str1 str2󕶎łȂCɓƂ trueԂ
proc ExistMatch {str1 str2} {
    if { [string match $str1 $str2] && [string match $str2 $str1] && [string length $str1] > 0 }  {
	return 1
    } else {
	return 0
    }
}


### OutputToSTDOUT ###
# ݑIĂ^OƂɃNĂ^Ȍo͂
proc OutputToSTDOUT {} {
    global tagdata oodata
    set record [${oodata}::current_record]
    ${record}::show_current_tag stdout
}


### CalcAscii ###
# Jn^OƏI^OdȂꍇC^O̕Ń\[g邽߂̐ԂD
# 擪 3lĂD
proc CalcAscii {in_string} {
    set ret_val ""

    for {set i 0} {$i < 3} {incr i} {
#  	puts [scan [string index $in_string $i] "%c"]
	set c_in_ascii [scan [string index $in_string $i] "%c"]
	if { [string match $c_in_ascii ""] } {
	    set ret_val [format "%s00000" $ret_val]
	} else {
	    set ret_val [format "%s%05d" $ret_val $c_in_ascii]
	}
    }
    return [expr 1000000000000000 - [string trimleft $ret_val "0"]]
}

### Message ###
proc Message {message} {
    global w
    tk_messageBox -title "Message" -message $message -type ok
}


### YN_Message ###
proc YN_Message {message} {
    global w
    return [tk_messageBox -title "Message" -message $message -type yesno]
}



### ResetText ###
# ݂̃eLXg̓eƃ^ÖʒuXVD
proc ResetText {} {
    global oodata w

    if {$w(edit-mode) == 0} {
	Message "Please change the mode to \"Edit mode\"."
 	Reflesh
	return 1
    }

    SetUndoData

    set record [${oodata}::current_record]
    ${record}::reset_text
}


proc Unfocus {} {
    global w

    focus $w(left)
    set w(unfocus) 1
    SetData
}


# -------------------- Bind --------------------
### Set binding ###
proc SetBinding {} {
    global w screens binding

    foreach i { OpenFile Save Export Reflesh Undo ISearch TSearch ResetText ToggleEditMode Unfocus exit } {
	bind all <$binding($i)> $i
    }

    foreach screen $screens {
	bind $screen <$binding(RemoveTag)> "RemoveTag key"
	bind $screen <Tab> NextNewData
	bind $screen <<PrevWindow>> PreviousNewData
	#     bind [set $screen] <Double-Button-1> ""
    }

    bind all <Tab> NextNewData
    bind all <<PrevWindow>> PreviousNewData

    bind all <Button-5> Next
    bind all <Button-4> Previous

    bind all <Next> Next
    bind all <Prior> Previous
}

### Call Main ###
Main

